"""
Generate samples of synthetic datasets.
Mainly for benchmarks and experiments
"""

import numpy as np
import pandas as pd


def make_transactions(
    n_transactions=1000, n_items=100, density=0.5, random_state=None, item_start=0
):
    """
    Generate a transactional dataset with predefined properties

    see: https://liris.cnrs.fr/Documents/Liris-3716.pdf

    Transaction sizes follow a normal distribution, centered around ``density * n_items``.
    Individual items are integer values between 0 and ``n_items``.

    Parameters
    ---------
    n_transactions: int, default=1000
        The number of transactions to generate
    n_items: int, default=100
        The number of indidual items, i.e the size of the set of symbols
    density: float, default=0.5
        Density of the resulting dataset
    random_state : int, RandomState instance, default=None
        Determines random number generation for dataset creation. Pass an int
        for reproducible output across multiple function calls.

    References
    ----------
    .. [1] F. Flouvat, F. De Marchi, JM. Petit
           "A new classification of datasets for frequent itemsets", 2009

    Example
    -------
    >>> from skmine.datasets import make_transactions
    >>> make_transactions(n_transactions=5, n_items=20, density=.25)  # doctest: +SKIP
    0    [0, 6, 18, 10, 1, 12]
    1          [2, 18, 10, 14]
    2                [4, 5, 1]
    3         [10, 11, 16, 19]
    4     [9, 4, 19, 8, 12, 5]
    dtype: object

    Notes
    -----
    With a binary matrix representation of the resulting dataset, we have the following equality
        .. math:: density = { Number\ of\ ones \over Number\ of\ cells }
    This is equivalent to
        .. math:: density = { Average\ transaction\ size \over number\ of\ items }

    Returns
    -------
    pd.Series: a Series of shape (``n_transactions``,)
        Earch entry is a list of integer values
    """
    if not 0.0 < density < 1.0:
        raise ValueError("density should be a float value between 0 and 1")

    avg_transaction_size = density * n_items

    generator = np.random.RandomState(random_state)  # pylint: disable= no-member

    item_stop = item_start + n_items
    choices = np.arange(start=item_start, stop=item_stop)
    t_sizes = generator.binomial(
        n=avg_transaction_size * 2,
        p=0.5,  # centered around avg_transaction_size
        size=n_transactions,
    )
    max_size = t_sizes.max()
    if max_size > n_items:
        delta = max_size - n_items
        t_sizes = np.clip(t_sizes, a_min=t_sizes.min() + delta, a_max=n_items)

    D = [generator.choice(choices, size, replace=False) for size in t_sizes]
    return pd.Series(D)


def make_classification(
    n_samples=100,
    n_items_per_class=100,
    *,  # pylint: disable= too-many-locals
    n_classes=2,
    weights=None,
    class_sep=0.2,
    shuffle=True,
    random_state=None,
    densities=None
):
    """
    Generate a random n-class classification problem

    Acts like sklearn version of make_classification, but produces
    transactional data instead. Transactions are drawn from a ``n_items_per_class``
    number of items, respecting the ``class_sep`` parameter to ensure transactions
    are drawn from different alphabets for different classes.

    A ``class_sep`` value of 0.0 will result in transactions being drawn from the
    same set of symbols.

    Densities can be defined for each class given the ``densities``
    parameter.

    Parameters
    ----------
    n_samples: int, default=100
        The number of samples
    n_items_per_class: int, default=100
        The number of items per class. This is similar to the ``n_features``
        parameters in scikit-learn, but operates at a class level.
    n_classes: int, default=2
        The number of classes (or labels) of the classification problem
    weigths, array-like of shape (n_classes,) default=None
        The proportions of samples assigned to each class. If None, then classes are balanced
    class_sep: float, default=0.2
        The factor of different items in different between classes.
        Setting this to 1.0 will make classification dummy.
    shuffle: boolean, default=True
        Shuffle the samples and the labels
    random_state: int RandomState instance, default=None
        Determines random number generation for dataset creation. Pass an int
        for reproducible output across multiple function calls.

    Returns
    -------
    D: pd.Series of shape [n_samples, ]
        The generated samples
    y: pd.Series of shape [n_samples]
        Labels associated to D

    See also
    --------
    make_transactions : which is used internally to generate samples
    """
    assert n_classes > 0
    if densities is None:
        densities = [0.5] * n_classes

    if weights is None:
        weights = [1 / n_classes] * n_classes  # balanced by default

    assert len(weights) == len(densities) == n_classes
    assert 0 <= class_sep <= 1.0
    np.testing.assert_almost_equal(np.sum(weights), 1.0, decimal=2)

    res = dict()

    padding = 0
    for _class in range(n_classes):
        _n_samples = int(weights[_class] * n_samples)
        density = densities[_class]
        transactions = make_transactions(
            n_transactions=_n_samples,
            n_items=n_items_per_class,
            random_state=random_state,
            item_start=padding,
            density=density,
        )
        res[_class] = transactions

        # if class_sep == 1.0, then separation is strict
        padding += int(n_items_per_class - (n_items_per_class * (1 - class_sep)))

    dfs = list()
    for _class, transactions in res.items():
        df = transactions.to_frame(name="transaction")
        df.loc[:, "class"] = _class
        dfs.append(df)

    df = pd.concat(dfs, axis=0)

    if shuffle:
        df = df.sample(frac=1)

    return df["transaction"], df["class"]



"""
Base IO for all periodic datasets
"""
import os
import warnings

import pandas as pd

from ._base import get_data_home


def deduplicate(S):
    """deduplicate pd.Series by removing rows with same index and values"""
    dedup = S.groupby(S.index).first()
    diff = len(S) - len(dedup)
    if diff:
        warnings.warn(f"found {diff} duplicates in S, removing them")
        return dedup
    return S


def fetch_health_app(data_home=None, filename="health_app.csv"):
    """Fetch and return the health app log dataset

    see: https://github.com/logpai/loghub

    HealthApp is a mobile application for Android devices.
    Logs were collected from an Android smartphone after 10+ days of use.

    Logs have been grouped by their types, hence resulting
    in only 20 different events.

    ==============================      ===================================
    Number of events                    20
    Average delta per event             Timedelta('0 days 00:53:24.984000')
    Average nb of points per event      100.0
    ==============================      ===================================

    Parameters
    ----------
    data_home : optional, default: None
        Specify another download and cache folder for the datasets. By default
        all scikit-mine data is stored in `scikit-mine_data`.

    Returns
    -------
    pd.Series
        System logs from the health app dataset, as an in-memory pandas Series.
        Events are indexed by timestamps.
    """
    data_home = data_home or get_data_home()
    p = os.path.join(data_home, filename)
    kwargs = dict(header=None, index_col=0, squeeze=True, dtype="string")
    if filename in os.listdir(data_home):
        s = pd.read_csv(p, **kwargs)
    else:
        s = pd.read_csv(
            "https://raw.githubusercontent.com/logpai/loghub/master/HealthApp/HealthApp_2k.log",
            sep="|",
            error_bad_lines=False,
            usecols=[0, 1],
            **kwargs,
        )
        s.to_csv(p, header=False)
    s.index.name = "timestamp"
    s.index = pd.to_datetime(s.index, format="%Y%m%d-%H:%M:%S:%f")

    return s


def fetch_canadian_tv(data_home=None, filename="canadian_tv.txt"):
    """
    Fetch and return canadian TV logs from August 2020

    see: https://zenodo.org/record/4671512

    If the dataset has never been downloaded before, it will be downloaded and stored.

    The returned dataset contains only TV series programs indexed by their associated timestamps.
    Adverts are ignored when loading the dataset.

    ==============================      =======================================
    Number of events                    98
    Average delta per event             Timedelta('19 days 02:13:36.122448979')
    Average nb of points per event      21.35714285714285
    ==============================      =======================================

    Parameters
    ----------
    data_home : optional, default: None
        Specify another download and cache folder for the datasets. By default
        all scikit-mine data is stored in `scikit-mine_data`.

    Returns
    -------
    pd.Series
        TV series events from canadian TV, as an in-memory pandas Series.
        Events are indexed by timestamps.

    Notes
    -----
    For now the entire .zip file is downloaded, being ~90mb on disk
    Downloading preprocessd dataset from zenodo.org is something we consider.

    See Also
    -------
    skmine.datasets.get_data_home
    """
    data_home = data_home or get_data_home()
    p = os.path.join(data_home, filename)
    kwargs = dict(header=None, squeeze=True, dtype="string", index_col=0,)

    if filename not in os.listdir(data_home):
        s = pd.read_csv(
            "https://zenodo.org/record/4671512/files/canadian_tv.txt", **kwargs
        )
        s.to_csv(p, index=True, header=False)
    else:
        s = pd.read_csv(p, **kwargs)

    s.index = pd.to_datetime(s.index)
    s.index.name = "timestamp"
    s.name = "canadian_tv"
    return s



"""  Function Utilities  """
import logging
import math
from collections import Counter

import networkx as nx
from networkx import Graph
from networkx.algorithms import isomorphism as iso

from skmine.graph.graphmdl.candidate import Candidate


def log2(value, total):
    """ Compute logarithm in the binary base
    Parameters
    ----------
    value
    total

    Returns
    -------
    float
    """
    return -math.log2(value / total)


def prequential_code(seq, epsilon=0.5):
    """ Compute the prequential code for a given sequence according an special epsilon
    Parameters
    -----------
    seq
    epsilon
    Returns
    -------
    float
    """
    code = 0.0
    eps = epsilon
    total_epsilon = len(seq) * epsilon
    for value in seq:
        for i in range(0, value):
            code += log2(eps, total_epsilon)
            total_epsilon += 1
            eps += 1

        eps = epsilon
    return code


def count_edge_label(graph: Graph):
    """ Count each edge label occurrence in a graph and store the result
    Parameters
    ----------
    graph:Graph
        the data graph
    Returns
    -------
    dict
    """

    edges = dict()
    j = 1
    for u, v, d in graph.edges(data=True):
        if 'label' in d:
            if type(d['label']) is tuple:
                for i in d['label']:
                    edges[j] = i
                    j = j + 1
            else:
                edges[j] = d['label']
                j += 1
    return dict([(i, j) for i, j in Counter(edges.values()).items()])


def count_vertex_label(graph: Graph):
    """ Count each vertex label occurrence in a graph and store the result
        Parameters
        ----------
        graph:Graph
            the data graph
        Returns
        -------
        dict
      """
    vertex = dict()
    j = 1
    for u, d in graph.nodes(data=True):
        if 'label' in d:
            if type(d['label']) is tuple:
                for i in d['label']:
                    vertex[j] = i
                    j = j + 1
            else:
                vertex[j] = d['label']
                j += 1
    return dict([(i, j) for i, j in Counter(vertex.values()).items()])


def get_total_label(graph: Graph):
    """ Compute total number of labels in the graph
       Parameters
       ----------
       graph:Graph
         the data graph
       Returns
       -------
         float
    """

    labels = count_edge_label(graph)
    labels.update(count_vertex_label(graph))
    total = 0.0  # The total number of labels in the graph
    for i in labels.values():
        total = total + i

    return total


def binomial(n, k):
    """ Compute the binomial coefficient for a given n and given k. Also called "n choose k"
    Parameters
    ----------
    n
    k
    Returns
    -------
    float
    """
    if k > n:
        raise ValueError(f"{k} should be lower than {n} in binomial coefficient")
    elif k == 0:
        return 1
    elif k > n / 2:
        return binomial(n, n - k)

    return n * binomial(n - 1, k - 1) / k


def universal_integer_encoding(x):
    """ Compute universal codeword sets and representation for integer except 1
    Parameters
    ----------
    x
    Returns
    -------
    int
    """
    if x < 1:
        raise ValueError(f"{x} should be higher than 1")
    else:
        return math.floor(math.log2(x)) + 2 * math.floor(math.log2(math.floor(math.log2(x)) + 1)) + 1


def universal_integer_encoding_with0(x):
    """ Compute universal codeword sets and representation for integer
    Parameters
    ----------
    x
    Returns
    -------
    int
    """
    if x < 0:
        raise ValueError(f"{x} should be higher than 0")
    else:
        return universal_integer_encoding(x + 1)


def encode(pattern: Graph, standard_table):
    """ Compute a given graph description length according a given label codes
    Parameters
    ----------
    standard_table : The label codes
    pattern : Graph
        The given graph
    Returns
    -------
    float
    """
    edges = count_edge_label(pattern)  # count each pattern edge label occurrences
    vertex = count_vertex_label(pattern)  # count each pattern vertex label occurrences

    # Get total number of label in the standard table
    total_label = len(standard_table.vertex_lc()) + len(standard_table.edges_lc())
    vertex_number = len(pattern.nodes())

    total_label_description = math.log2(total_label)  # description length for all labels
    vertex_number_description = universal_integer_encoding_with0(vertex_number)  # description length for all vertex

    # Compute description length for vertex
    vertex_description = dict()
    for u, v in vertex.items():
        desc = standard_table.vertex_lc()[u] + universal_integer_encoding_with0(v) + math.log2(
            binomial(vertex_number, v))
        vertex_description[u] = desc

    # Compute description length for edges
    edges_description = dict()
    for a, b in edges.items():
        desc = standard_table.edges_lc()[a] + universal_integer_encoding_with0(b) + math.log2(
            binomial(int(math.pow(vertex_number, 2)), b))
        edges_description[a] = desc

    # Compute description length through description length of edges and vertex
    description_length = 0.0
    for i in vertex_description.values():
        description_length = description_length + i
    for j in edges_description.values():
        description_length = description_length + j

    return description_length + total_label_description + vertex_number_description


def encode_vertex_singleton(standard_table, vertex_label):
    """ Compute a vertex singleton description length
        Parameters
        ----------
        standard_table : a labels codes
        vertex_label : The vertex singleton label
        Returns
        -------
        float
    """
    if vertex_label == "" or vertex_label is None:
        raise ValueError(f"You should give a vertex label")
    else:
        # Get total number of label in the standard table
        total_label = len(standard_table.vertex_lc()) + len(standard_table.edges_lc())
        total_label_description = math.log2(total_label)  # description length for all labels
        vertex_number_description = universal_integer_encoding_with0(1)  # description length for all vertex

        # Compute description length for vertex
        desc = standard_table.vertex_lc()[vertex_label] + universal_integer_encoding_with0(1) + math.log2(
            binomial(1, 1))

        return desc + total_label_description + vertex_number_description


def encode_edge_singleton(standard_table, edge_label):
    """ Compute an edge singleton description length
        Parameters
        ----------
        standard_table : a labels codes
        edge_label : the edge singleton label
        Returns
        -------
        float
    """
    if edge_label == "" or edge_label is None:
        raise ValueError("You should give an edge label")
    else:
        # Get total number of label in the standard table
        total_label = len(standard_table.vertex_lc()) + len(standard_table.edges_lc())
        total_label_description = math.log2(total_label)  # description length for all labels
        vertex_number_description = universal_integer_encoding_with0(2)  # description length for all vertex

        # Compute description length for vertex
        desc = standard_table.edges_lc()[edge_label] + universal_integer_encoding_with0(1) + math.log2(binomial(4, 1))

        return desc + total_label_description + vertex_number_description


def encode_singleton(standard_table, arity, label):
    """ Compute a singleton description length according her arity
        Parameters
        ----------
        standard_table : a label codes
        label : The singleton label
        arity : 1 for the vertex singleton and 2 for the edge singleton
        Returns
        -------
        float
    """

    if arity == 1:
        return encode_vertex_singleton(standard_table, label)
    elif arity == 2:
        return encode_edge_singleton(standard_table, label)
    else:
        raise ValueError("arity should must be 1 or 2")


def _node_match(node1, node2):
    """ Compare two given nodes
    Parameters
    ---------
    node1
    node2
    Returns
    -------
    bool
    """
    if 'label' in node1 and 'label' in node2:
        res = list()
        if type(node1['label']) is str and type(node2['label']) is str:
            res.append(node1['label'] == node2['label'])
        elif type(node1['label']) is not str and type(node2['label']) is str:
            res.append(node2['label'] in node1['label'])
        elif type(node1['label']) is not str and type(node2['label']) is not str:
            for i in node2['label']:
                res.append(i in node1['label'])
        else:
            res.append(False)

        return not (False in res)
    elif 'label' not in node1 and 'label' in node2:
        return False
    else:
        return True


def _edge_match(edge1, edge2):
    """ Compare two given edges
    Parameters
    ----------
    edge1
    edge2
    Returns
    -------
    bool
    """
    if 'label' in edge1 and 'label' in edge2:
        return edge1['label'] == edge2['label']
    else:
        return ValueError("All edges must have labels")


def _edge_match_for_multigraph(edge1, edge2):
    """ Compare two given nodes in a given multigraph
    Parameters
    ----------
    edge1
    edge2
    Returns
    -------
    bool
    """
    edge1_values = dict(edge1).values()
    edge2_values = dict(edge2).values()
    edge1_labels = [list(edge1_values)[0][i] for i in list(edge1_values)[0].keys() if i == 'label']
    edge2_labels = [list(edge2_values)[0][i] for i in list(edge2_values)[0].keys() if i == 'label']
    if len(edge2_labels) != 0 and len(edge1_labels) != 0:
        return not (False in [label in edge1_labels for label in edge2_labels])
    else:
        raise ValueError("All edges must have labels")


def get_embeddings(pattern, graph):
    """ Provide the embeddings of a pattern in a given graph
    Parameters
    ----------
    pattern
    graph
    Returns
    -------
    list
    """

    # Create functions to compare node and edge label
    comp = {
        'node_match': _node_match,
        'edge_match': _edge_match
    }

    comp_multigraph = {
        'node_match': _node_match,
        'edge_match': _edge_match_for_multigraph
    }
    graph_matcher = None
    # Create matcher according the graph type (directed or no)
    if type(graph) is nx.DiGraph:
        graph_matcher = iso.DiGraphMatcher(graph, pattern, **comp)
    elif type(graph) is nx.MultiDiGraph:
        graph_matcher = iso.MultiDiGraphMatcher(graph, pattern, **comp_multigraph)
    elif type(graph) is nx.MultiGraph:
        graph_matcher = iso.MultiGraphMatcher(graph, pattern, **comp_multigraph)
    else:
        graph_matcher = iso.GraphMatcher(graph, pattern, **comp)
    return list(graph_matcher.subgraph_monomorphisms_iter())


def is_vertex_singleton(pattern):
    """ Check if a given pattern is a vertex singleton pattern
    Parameters
    ---------
    pattern
    Returns
    ---------
    bool
    """
    return len(pattern.nodes()) == 1 and get_total_label(pattern) == 1


def is_edge_singleton(pattern):
    """ Check if a given pattern is an edge singleton pattern
    Parameters
    ---------
    pattern
    Returns
    ---------
    bool
    """
    if len(pattern.nodes()) == 2:
        # Check first if the nodes haven't labels
        if "label" not in pattern.nodes(data=True)[1] \
                and "label" not in pattern.nodes(data=True)[2]:
            # Check if the edge have exactly one label
            if count_edge_label(pattern) is not None \
                    and len(count_edge_label(pattern).values()) == 1 \
                    and list(count_edge_label(pattern).values())[0] == 1:
                return True
            else:
                return False
        else:
            return False
    else:
        return False


def get_support(embeddings):
    """ Compute the pattern support in the graph according a minimum image based support

    The minimum image-based support is the minimum number of graph nodes,
    that a pattern node can replace

    Parameters
    ---------
    embeddings
    Returns
    -------
    int
    """
    if len(embeddings) != 0:
        node_embeddings = dict()
        # Compute for each pattern node, the graph nodes who can replace,
        # and store it in a dictionary
        for e in embeddings:
            for i in e.items():
                if i[1] in node_embeddings:
                    node_embeddings[i[1]].add(i[0])
                else:
                    node_embeddings[i[1]] = set()
        embed = dict()
        # Compute for each pattern node,the total number of graph nodes who could replace,
        # and store it in a dictionary
        for key, value in node_embeddings.items():
            embed[key] = len(value)

        return min(embed.values())
    else:
        return 0


def get_label_index(label, values):
    """ Provide a label index in the values
    Parameters
    ----------
    label
    values
    Returns
    ----------
    int
    """
    if label in values:
        return values.index(label)
    else:
        raise ValueError(f"{label} should be in the {values}")


def get_node_label(key, index, graph):
    """ Provide a particular node label in a given graph by the index and the node key
    Parameters
    ----------
    key
    index
    graph
    Returns
    ---------
    str
    """
    if key in graph.nodes():
        if len(graph.nodes(data=True)[key]['label']) > index and type(graph.nodes(data=True)[key]['label']) is tuple:
            return graph.nodes(data=True)[key]['label'][index]
        else:
            return graph.nodes(data=True)[key]['label']
    else:
        raise ValueError(f"{index} shouldn't be out of bounds and {key} should be a graph node")


def get_edge_label(start, end, graph):
    """ Provide a particular edge label in a given graph by the edge start and the edge end
    Parameters
    ---------
    start
    end
    graph
    Returns
    ---------
    str
    """
    if start in graph.nodes() and end in graph.nodes():
        if type(graph) is nx.MultiDiGraph:
            if (start, end) in list(graph.edges(start)):
                if not (False in ['label' in v
                                  for v in graph.get_edge_data(start, end).values()]):
                    return [v['label'] for v in graph.get_edge_data(start, end).values()]
            elif (end, start) in list(graph.edges(end)):
                if not (False in ['label' in v
                                  for v in graph.get_edge_data(end, start).values()]):
                    return [v['label'] for v in graph.get_edge_data(end, start).values()]
            else:
                raise ValueError(f"{start}-{end} should be a graph edge and should have a label")
        else:
            if (start, end) in list(graph.edges(start)):
                if 'label' in graph[start][end]:
                    return graph[start][end]['label']
            elif (end, start) in list(graph.edges(end)):
                if 'label' in graph[end][start]:
                    return graph[end][start]['label']
            else:
                raise ValueError(f"{start}-{end} should be a graph edge and should have a label")
    else:
        raise ValueError(f"{start} and {end} should be a graph nodes")


def is_without_edge(pattern):
    """ Check if the pattern is without edge
    Parameters
    ----------
    pattern
    Returns
    ---------
    bool
    """
    return len(pattern.edges()) == 0


def _get_node_labels(node, graph):
    """ Provide the labels for a particular graph node,
    if the node doesn't have labeled, return the node

    Parameters
    ----------
    node
    graph
    Returns
    --------
    str
    """
    if 'label' in graph.nodes[node]:
        return graph.nodes[node]['label']
    else:
        return node


def display_graph(graph: Graph):
    """ Display a given graph in a specific string sequence
    Parameters
    ---------
    graph
    Returns
    --------
    str
    """
    msg = ""
    if len(graph.edges()) != 0:
        for edge in graph.edges(data=True):
            if "label" in edge[2]:
                msg += "{}--{}-->{}".format(_get_node_labels(edge[0], graph),
                                            edge[2]['label'], _get_node_labels(edge[1], graph)) + "\n"
            else:
                msg += "{}--->{}".format(_get_node_labels(edge[0], graph),
                                         _get_node_labels(edge[1], graph)) + "\n"
    else:
        msg += "{}".format(_get_node_labels(1, graph))
    return msg


def draw_graph(g: Graph):
    """ Draw a given pattern
    Parameters
    ----------
    g
    """
    pos = nx.spring_layout(g, seed=7)
    edge_labels = dict([((u, v,), d['label']) for u, v, d in g.edges(data=True)])
    node_labels = dict([(u, d['label']) for u, d in g.nodes(data=True) if 'label' in d])
    nx.draw(g, pos, with_labels=True)
    nx.draw_networkx_edge_labels(g, pos, edge_labels=edge_labels, font_color="red", font_weight="bold", font_size=14)
    nx.draw_networkx_labels(g, pos, node_labels, font_color="red", font_weight="bold", font_size=14)


def get_edge_in_embedding(embedding, pattern):
    """ Provide the pattern edges who are in a given embedding
    Parameters
    ----------
    embedding
    pattern
    Returns
    -------
    set
    """
    keys = list(embedding.keys())
    values = list(embedding.values())
    edges = set()
    i = 0
    while i <= len(keys) - 1:
        j = i
        pattern_edges = list(pattern.edges())
        while j <= len(keys) - 1:
            if (values[i], values[j]) in pattern_edges:
                edges.add((values[i], values[j]))
            if (values[j], values[i]) in pattern_edges:
                edges.add((values[j], values[i]))
            j += 1
        del pattern_edges
        i += 1
    return edges


def get_key_from_value(data, value):
    """ Provide a dictionary key from the value
    Parameters
    ----------
    data
    value
    Returns
    --------
    int
    """
    return [k for k, v in data.items() if v == value][0]


def get_two_nodes_all_port(node1, node2, rewritten_graph):
    """ Provide all port in a graph for a potential candidate nodes
     Parameters
     ----------
     node1
     node2
     rewritten_graph
     Returns
     -------
     list
     """
    # check for all port edges if the node 1 and the node2 are port neighbors
    res = set()
    for node in rewritten_graph.nodes(data=True):
        if 'is_Pattern' not in node[1]:
            edge1 = (node1, node[0])
            edge2 = (node2, node[0])
            cpt = 0
            for edge in rewritten_graph.in_edges(node[0]):
                if edge == edge1 or edge == edge2:
                    cpt += 1
            if cpt == 2:
                res.add(node[0])
    return list(res)


def get_all_candidate_ports_labels_tuple(rewritten_graph, ports, first_node, second_node, inverse=False):
    """ Provide a port tuples for a given two nodes who are a candidate
    Parameters
    ---------
    rewritten_graph
    ports : a complete list of data port
    first_node
    second_node
    inverse : boolean to decide for the tuple elements positions

    Returns
    -------
    list"""
    port = []
    for p in ports:
        first_port = get_edge_label(first_node, p, rewritten_graph)
        second_port = get_edge_label(second_node, p, rewritten_graph)
        if not inverse:
            port.append((first_port, second_port))
        else:
            port.append((second_port, first_port))
    return port


def is_isomorphic(graph, pattern):
    """ Check if two graph is isomorphic
    Parameters
    ---------
    graph
    pattern
    Returns
    -------
    bool
    """

    opt = {
        'node_match': _node_match,
        'edge_match': _edge_match
    }
    graph_matcher = None
    # Create matcher according the graph type (directed or no)

    if nx.is_directed(graph):
        graph_matcher = iso.DiGraphMatcher(graph, pattern, **opt)
    else:
        graph_matcher = iso.GraphMatcher(graph, pattern, **opt)

    return graph_matcher.is_isomorphic()


def get_automorphisms(graph):
    """ Provide a given graph automorphisms
    Parameters
    ----------
    graph
    Returns
    ---------
    list
    """
    opt = {
        'node_match': _node_match,
        'edge_match': _edge_match
    }
    graph_matcher = None
    # Create matcher according the graph type (directed or no)

    if nx.is_directed(graph):
        graph_matcher = iso.DiGraphMatcher(graph, graph, **opt)
    else:
        graph_matcher = iso.GraphMatcher(graph, graph, **opt)

    """automorphisms = set()
    for auto in list(graph_matcher.isomorphisms_iter())[1:]:
        for i, j in auto.items():
            if i != j:
                automorphisms.add((i, j))"""

    return list(graph_matcher.isomorphisms_iter())


def get_port_candidates(patterns_list):
    """ Provide a list of node tuple from the given list
    Parameters
    ----------
    patterns_list
    Returns
    -------
    list
    """
    res = set()
    i = 0
    while i <= len(patterns_list) - 1:
        j = i + 1
        while j <= len(patterns_list) - 1:
            res.add((patterns_list[i], patterns_list[j]))
            j += 1
        i += 1
    return list(res)


def compute_pattern_usage(rewritten_graph, pattern_format, ports):
    """ Compute pattern usage in the rewritten graph
    Parameters
    ---------
    rewritten_graph
    pattern_format: The pattern label concatenated with the port label
    ports : The pattern candidates ports
    Returns
    -------
    float
    """
    pattern_usage = 0
    ports_infos = get_port_node_infos(rewritten_graph)
    # compute for each given ports, the information about his pattern neighbors
    for port in ports:
        infos = ports_infos[port]
        # compute how many times, the pattern is the port neighbors
        for i in infos:
            if i in pattern_format:
                pattern_usage += 1
        # If the pattern have different appearances as neighbors,
        # return the pattern total usage sliced by the number of appearances
    return pattern_usage / len(pattern_format)


def compute_candidate_usage(rewritten_graph, candidate, code_table, candidates):
    """ Compute an estimated usage for a  given candidate
    Parameters
    -----------
    rewritten_graph
    candidate
    code_table : a Code Table
    candidates :candidates list
    """
    p1_format = []  # format of the first candidate pattern with its ports
    p2_format = []  # format of the second candidate pattern with its ports
    for p in candidate.port:
        p1_format.append(candidate.first_pattern_label + p[0])
        p2_format.append(candidate.second_pattern_label + p[1])

    # Compute each candidate patterns usage
    first_pattern_usage = compute_pattern_usage(rewritten_graph, p1_format, list(candidate.data_port))
    second_pattern_usage = compute_pattern_usage(rewritten_graph, p2_format, list(candidate.data_port))

    # Compute the candidate estimated usage according the candidate particularities
    if candidate.first_pattern_label == candidate.second_pattern_label and \
            (candidate == candidate.inverse() or candidate.inverse() not in candidates):
        candidate.set_usage(int(first_pattern_usage / 2))
    elif candidate.first_pattern is not None and candidate.second_pattern is not None:
        if is_without_edge(candidate.first_pattern) and not is_without_edge(candidate.second_pattern):
            # if only the second pattern doesn't have edges
            candidate.set_usage(second_pattern_usage)
        elif is_without_edge(candidate.second_pattern) and not is_without_edge(candidate.first_pattern):
            # if only the second pattern doesn't have edges
            candidate.set_usage(first_pattern_usage)
        elif is_without_edge(candidate.first_pattern) and is_without_edge(candidate.second_pattern):
            candidate.set_usage(compute_pattern_embeddings(rewritten_graph, candidate.first_pattern_label))
        else:
            # if both of the pattern have edges
            candidate.set_usage(min(first_pattern_usage, second_pattern_usage))

    elif candidate.second_pattern is not None and candidate.first_pattern is None:
        raise ValueError("The singleton should be the second pattern")
    elif candidate.first_pattern is not None and candidate.second_pattern is None:

        # A pattern with a singleton
        if code_table.is_ct_edge_singleton(candidate.second_pattern_label):
            # the singleton is an edge singleton
            if is_without_edge(candidate.first_pattern):
                # if the pattern doesn't have edge
                candidate.set_usage(second_pattern_usage)
            else:
                candidate.set_usage(min(first_pattern_usage, second_pattern_usage))
        else:
            # The singleton is a vertex singleton
            candidate.set_usage(first_pattern_usage)

    else:
        # Only singleton case
        if code_table.is_ct_edge_singleton(candidate.first_pattern_label) \
                and not code_table.is_ct_edge_singleton(candidate.second_pattern_label):
            # only first pattern is an edge singleton
            candidate.set_usage(first_pattern_usage)
        elif not code_table.is_ct_edge_singleton(candidate.first_pattern_label) \
                and code_table.is_ct_edge_singleton(candidate.second_pattern_label):
            # only second pattern is an edge singleton
            candidate.set_usage(second_pattern_usage)
        else:
            # both singleton is vertex singleton
            candidate.set_usage(min(first_pattern_usage, second_pattern_usage))


def compute_pattern_embeddings(rewritten_graph, pattern):
    """ Compute pattern embeddings in the rewritten graph
    Parameters
    ----------
    rewritten_graph
    pattern
    Returns
    --------
    int
    """
    res = 0
    for node in rewritten_graph.nodes(data=True):
        if node[1]['label'] == pattern:
            res += 1
    return res


def is_candidate_port_exclusive(candidates, candidate, port):
    """ Check if a given  port are neighbors who are not the candidate nodes number
     Parameters
     -----------
     candidates : Candidates list
     candidate
     port : The given port
     Returns
     ----------
     bool
     """
    res = []
    # search in the candidates list if there is one
    # who are the given candidate data port as his data port
    for c in candidates:
        for p in c.data_port:
            if p == port:
                res.append(c == candidate)

    return not (False in res)


def generate_candidates(rewritten_graph, code_table):
    """ Search in the rewritten graph, the pattern who share a same port
    Parameters
    ----------
    rewritten_graph
    code_table
    Returns
    ----------
    list
    """
    candidates = list()
    ports = [node[0] for node in rewritten_graph.nodes(data=True) if 'is_Pattern' not in node[1]]
    for port in ports:
        patterns = [edge[0] for edge in rewritten_graph.in_edges(port)]  # the port neighbor
        port_candidates = get_port_candidates(patterns)
        # del patterns
        for c in port_candidates:
            first_pattern = rewritten_graph.nodes[c[0]]
            second_pattern = rewritten_graph.nodes[c[1]]
            all_candidate_port = get_two_nodes_all_port(c[0], c[1], rewritten_graph)

            # respect the candidate pattern order

            # If both candidate element are the pattern
            if 'is_singleton' not in first_pattern and 'is_singleton' not in second_pattern:

                if int(first_pattern['label'].split('P')[1]) < int(second_pattern['label'].split('P')[1]):
                    ports = get_all_candidate_ports_labels_tuple(rewritten_graph, all_candidate_port, c[0], c[1])
                    candidate = Candidate(first_pattern['label'], second_pattern['label'], ports)
                else:
                    ports = get_all_candidate_ports_labels_tuple(rewritten_graph, all_candidate_port, c[0], c[1], True)
                    candidate = Candidate(second_pattern['label'], first_pattern['label'], ports)

                candidate.first_pattern = code_table.rows()[int(candidate.first_pattern_label.split('P')[1])].pattern()
                candidate.second_pattern = code_table.rows()[
                    int(candidate.second_pattern_label.split('P')[1])].pattern()

            # if one node is a pattern and the second a singleton
            elif 'is_singleton' in first_pattern and 'is_singleton' not in second_pattern:
                ports = get_all_candidate_ports_labels_tuple(rewritten_graph, all_candidate_port, c[0], c[1], True)
                candidate = Candidate(second_pattern['label'], first_pattern['label'], ports)
                candidate.first_pattern = code_table.rows()[int(candidate.first_pattern_label.split('P')[1])].pattern()
                # candidate.second_pattern = create_singleton_pattern(candidate.second_pattern_label, code_table)

            elif 'is_singleton' not in first_pattern and 'is_singleton' in second_pattern:
                ports = get_all_candidate_ports_labels_tuple(rewritten_graph, all_candidate_port, c[0],
                                                             c[1])
                candidate = Candidate(first_pattern['label'], second_pattern['label'], ports)
                candidate.first_pattern = code_table.rows()[int(candidate.first_pattern_label.split('P')[1])].pattern()
                candidate.second_pattern = create_singleton_pattern(candidate.second_pattern_label, code_table)

            # if both of the candidate pattern are singletons
            else:  # if both of the candidate elements are the singleton
                if first_pattern['label'] < second_pattern['label']:
                    ports = get_all_candidate_ports_labels_tuple(rewritten_graph, all_candidate_port, c[0], c[1])
                    candidate = Candidate(first_pattern['label'], second_pattern['label'], ports)
                else:
                    ports = get_all_candidate_ports_labels_tuple(rewritten_graph, all_candidate_port, c[0], c[1], True)
                    candidate = Candidate(second_pattern['label'], first_pattern['label'], ports)

            # Store all candidates' data_port to see if there are exclusive port
            for p in all_candidate_port:
                candidate.data_port.add(p)

            if candidate not in candidates:
                candidates.append(candidate)
            else:
                c = candidates[candidates.index(candidate)]
                for p in candidate.data_port:
                    c.data_port.add(p)

    for r in candidates:
        compute_candidate_usage(rewritten_graph, r, code_table, candidates)
        exclusive_port_number = 0
        # Search exclusive port
        for port in r.data_port:
            if is_candidate_port_exclusive(candidates, r, port):
                exclusive_port_number += 1
        r.exclusive_port_number = exclusive_port_number

        # create singleton pattern
        if r.first_pattern is None and r.second_pattern is not None:
            raise ValueError("The second pattern should be the singleton")
        elif r.first_pattern is not None and r.second_pattern is None:
            r.second_pattern = create_singleton_pattern(r.second_pattern_label, code_table)
        elif r.first_pattern is None and r.second_pattern is None:
            r.first_pattern = create_singleton_pattern(r.first_pattern_label, code_table)
            r.second_pattern = create_singleton_pattern(r.second_pattern_label, code_table)
    return candidates


def create_singleton_pattern(label, code_table):
    """ Create a graph who represent a singleton
    Parameters
    ----------
    label
    code_table
    Returns
    -------
    Graph
    """
    if code_table.data_is_multigraph():
        pattern = nx.MultiDiGraph()
    else:
        pattern = nx.DiGraph()

    if code_table.is_ct_edge_singleton(label):
        pattern.add_nodes_from(range(1, 3))
        pattern.add_edge(1, 2, label=label)
    elif code_table.is_ct_vertex_singleton(label):
        pattern.add_node(1, label=label)
    else:
        raise ValueError("The label should be a vertex or an edge label")
    return pattern


def count_port_node(rewritten_graph):
    """ Count the port number in a rewritten graph
    Parameters
    ----------
    rewritten_graph
    Returns
    --------
    int
    """
    numb = 0
    for node in rewritten_graph.nodes(data=True):
        if 'is_Pattern' not in node[1]:
            numb += 1
    return numb


def get_pattern_node_infos(rewritten_graph):
    """ Provide pattern node information from the rewritten graph
    Parameters
    ----------
    rewritten_graph
    Returns
    --------
    dict
    """
    pattern_node = dict()
    for node in rewritten_graph.nodes(data=True):
        if 'is_Pattern' in node[1]:
            res = []
            for edge in rewritten_graph.edges(node[0], data=True):
                res.append(edge[2]['label'])

        if node[1]['label'] in pattern_node:
            pattern_node[node[1]['label']].append(res)
        else:
            pattern_node[node[1]['label']] = []
            pattern_node[node[1]['label']].append(res)

        if 'is_singleton' in node[1]:
            if node[1]['is_singleton'] is True:
                if 'singleton' not in pattern_node[node[1]['label']]:
                    pattern_node[node[1]['label']].append('singleton')
            else:
                raise ValueError("is_singleton should be true or shouldn't exist")

    return pattern_node


def get_port_node_infos(rewritten_graph):
    """ Provide port node information from the rewritten graph
    Parameters
    ----------
    rewritten_graph
    Returns
    -------
    dict
    """
    port_node = dict()
    for node in rewritten_graph.nodes(data=True):
        if 'is_Pattern' not in node[1]:
            for edge in rewritten_graph.in_edges(node[0], data=True):
                if node[0] in port_node:
                    port_node[node[0]].append(rewritten_graph.nodes(data=True)[edge[0]]['label'] + edge[2]['label'])
                else:
                    port_node[node[0]] = []
                    port_node[node[0]].append(rewritten_graph.nodes(data=True)[edge[0]]['label'] + edge[2]['label'])
    return port_node


def get_port_node(rewritten_graph, node):
    """ Provide an embedding vertex port
    Parameters
    ----------
    rewritten_graph
    node
    Returns
    --------
    set
    """
    res = set()
    for edge in rewritten_graph.out_edges(node, data=True):
        res.add(int(edge[2]['label'].split('v')[1]))
    return res


def get_graph_from_file(file):
    """ Construct a networkx graph from a given file
    Parameters
    ----------
    file
    Returns
    -------
    Graph
    """
    file = open(file)
    lines = [line.split(' ') for line in file][3:]
    graph = nx.DiGraph()
    for line in lines:
        if line[0] == 'v':
            graph.add_node(int(line[1]) + 1, label=line[3].split('\n')[0])
        else:
            graph.add_edge(int(line[1]) + 1, int(line[2]) + 1, label=line[3].split('\n')[0])
    file.close()
    return graph


""" GraphMDL logger"""
MyLogger = logging.getLogger("skmine.graph.graphmdl")



"""
The :mod:`skmine.itemset` module includes methods which consider reccurent itemsets.
An itemset is an unsorted set of items, eg. identifiers for books in a store, words in a text, etc.
"""
from .lcm import LCM
from .slim import SLIM



import logging
import time
from collections import defaultdict
import numpy
import pandas
from .code_table_row import CodeTableRow
import math
import networkx as nx
import skmine.graph.graphmdl.utils as utils


class CodeTable:
    """
        Code table inspired from Krimp algorithm.

        It's composed by the non singleton pattern and their information (usage, ports) who represented by a row
        And the singleton pattern stored in a dictionary if there is one.

        The separation between singleton and non-singleton pattern make easily their treatments and reduces the execution time.

        Parameters
        ----------
        standard_table
            It's the data graph, label codes
        graph
            It's the data graph
    """

    def __init__(self, standard_table, graph):
        self._standard_table = standard_table  # A initial code table
        self._rows = []  # All rows of the code table
        self._description_length = 0.0  # the description length of this code table
        self._data_graph = graph  # The graph where we want to apply the code table elements
        """ We don't store singleton pattern ports because each singleton pattern node is its port,
            then the vertex singleton ports are {1:1}, node 1 and usage 1,
            and the edge singleton, {1:1, 2:1}, node 1, usage 1, node 2, usage 1.
            Thus singleton ports are intuitive 
        """
        self._vertex_singleton_usage = dict()  # singleton vertex usage
        self._edge_singleton_usage = dict()  # singleton edge usage
        self._singleton_used_embeddings = dict()  # singleton used embedding
        self._singleton_code_length = dict()  # singleton pattern code length
        self._rewritten_graph = nx.DiGraph()  # A directed graph who represents the rewritten graph

    def add_row(self, row: CodeTableRow):
        """ Add a new row at the code table
        Parameters
        ----------
        row
        """
        if row.arrival_number is None:
            row.arrival_number = len(self._rows) + 1

        self._rows.append(row)  # Add the new row at the end of the list

        # compute the row pattern embeddings and store them
        row.set_embeddings(utils.get_embeddings(row.pattern(), self._data_graph))

        # sort the list who contains the row according a reverse order and a specific key
        self._rows.sort(reverse=True, key=self._order_rows)

    def remove_row(self, row: CodeTableRow):
        """ Remove a row on the code table rows
        Parameters
        ---------
        row
        """
        # del self._rows[self.rows().index(row)]
        self._rows.remove(row)
        self._rows.sort(reverse=True, key=self._order_rows)

    def rows(self):
        """ Provide code table rows
        Returns
        -------
        list
        """
        return self._rows

    def cover(self):
        """ Make the cover for the code table,

            the cover marker is get from the data graph
        """
        b = time.time()
        # Get the cover marker and increment it
        if 'cover_marker' in self._data_graph.graph:
            self._data_graph.graph['cover_marker'] += 1
        else:
            self._data_graph.graph['cover_marker'] = 1

        cover_marker = self._data_graph.graph['cover_marker']  # the current cover marker

        self._rewritten_graph = nx.DiGraph()  # reset the rewritten graph before each cover

        for row in self._rows:
            self.row_cover(row, self._data_graph, cover_marker, self._rewritten_graph, self.rows().index(row))

        res = self.singleton_cover(self._data_graph, cover_marker, self._rewritten_graph)

        self._vertex_singleton_usage = res[0]  # Store vertex singleton usage
        self._edge_singleton_usage = res[1]  # Store edge singleton usage
        self._singleton_used_embeddings = res[2]  # store the used embeddings
        usage_sum = self._compute_usage_sum()  # Get the total of the rows usage
        # compute each row code length and description length
        for row in self._rows:
            row.compute_code_length(usage_sum)
            # row.compute_description_length(self._standard_table)

        self._compute_singleton_code(usage_sum)  # compute singleton code length
        utils.MyLogger.info(f"cover time...{time.time() - b}")

    def compute_total_description_length(self):
        """ Compute total description length according the prequential code equation
        Returns
        -------
        double
        """
        self.compute_ct_description_length()

        return self._description_length + self.compute_rewritten_graph_description()

    def compute_ct_description_length(self):
        """ Compute this code table description length
            according kgmdl equation (with prequential code) """
        if len(self._rewritten_graph.nodes()) != 0:  # Check if the cover is done
            # check if the code table is already covered
            description_length = 0.0
            for row in self._rows:
                if row.pattern_usage() != 0:
                    row.compute_description_length(self._standard_table)
                    description_length += row.description_length()

            description_length += self._compute_singleton_description_length()

            self._description_length = description_length
        else:
            raise ValueError("You should cover the code table before computing his description")

    def compute_rewritten_graph_description(self):
        """ Compute _description_length of the rewritten graph,
            according kgmdl equation
        Returns
        -------
        float
        """
        if len(self._rewritten_graph.nodes()) != 0:
            desc = 0.0
            desc += utils.universal_integer_encoding_with0(
                len([node for node in self._rewritten_graph.nodes(data=True)
                     if 'is_Pattern' in node[1]]))  # embedding vertices
            desc += math.log2(len([node for node in self._rewritten_graph.nodes(data=True)
                                   if 'is_Pattern' not in node[1]]) + 1)  # port vertices count
            # number of edges per vertex
            edges_per_vertex_seq = [
                row.pattern_usage() * math.log2(len(row.port_code_length().keys()) + 1)
                for row in self._rows if row.pattern_usage() != 0]
            edges_per_vertex_seq.extend([v * math.log2(1 + 1) for v in self._vertex_singleton_usage.values()])
            edges_per_vertex_seq.extend((v * math.log2(2 + 1) for v in self._edge_singleton_usage.values()))
            desc += sum(edges_per_vertex_seq)

            # pattern of each vertex embedding sequence
            pattern_embedding = [row.pattern_usage() for row in self._rows if row.pattern_usage() != 0]
            pattern_embedding.extend([v for v in self._vertex_singleton_usage.values()])
            pattern_embedding.extend((v for v in self._edge_singleton_usage.values()))
            desc += utils.prequential_code(pattern_embedding)

            # edges destination
            desc += utils.prequential_code(
                [len(self._rewritten_graph.in_edges(node[0]))
                 for node in self._rewritten_graph.nodes(data=True)
                 if 'is_Pattern' not in node[1]])

            # edges labels
            edges_label = [utils.prequential_code([v for v in row.pattern_port_usage().keys()])
                           for row in self._rows if row.pattern_usage() != 0]
            # the code for vertex pattern is zero
            edges_label.extend([utils.prequential_code([v, v]) for v in self._edge_singleton_usage.values()])
            desc += sum(edges_label)
            return desc
        else:
            raise ValueError("You should first cover the code table")

    def description_length(self):
        """ Provide the code table description length
        Returns
        -------
        float
        """
        return self._description_length

    def rewritten_graph(self):
        """ Provide the code table rewritten graph
        Returns
        -------
        object
        """
        return self._rewritten_graph

    def label_codes(self):
        """ Provide the label codes
        Returns
        -------
        object
        """
        return self._standard_table

    def is_ct_edge_singleton(self, label):
        """ provide if a given label is an edge singleton label
        Parameters
        ----------
        label
        Returns
        -------
        bool
        """
        return label in self._edge_singleton_usage

    def is_ct_vertex_singleton(self, label):
        """ provide if a given label is a vertex singleton label
        Parameters
        ----------
        label
        Returns
        -------
        bool
        """
        return label in self._vertex_singleton_usage

    def data_port(self):
        """ Provide all graph data port
        Returns
        -------
        list
        """
        data_port = []
        for node in self._data_graph.nodes(data=True):
            if 'port' in node[1]:
                data_port.append(node[0])

        return data_port

    def _compute_usage_sum(self):
        """ Compute the total of usage for this code table elements
        Returns
        -------
        float : the usage sum """
        usage_sum = 0.0
        for row in self._rows:
            usage_sum += row.pattern_usage()

        if len(self._vertex_singleton_usage.keys()) != 0:
            for value in self._vertex_singleton_usage.values():
                usage_sum += value

        if len(self._edge_singleton_usage.keys()) != 0:
            for value in self._edge_singleton_usage.values():
                usage_sum += value

        return usage_sum

    def _compute_singleton_code(self, usage_sum):
        """ Compute singleton pattern code length after cover
        Parameters
        ----------
        usage_sum : Total of pattern ( even singleton) usage in the code table
        """
        self._singleton_code_length = dict()
        if len(self._vertex_singleton_usage) != 0:
            for u, v in self._vertex_singleton_usage.items():
                if utils.log2(v, usage_sum) == -0.0:
                    self._singleton_code_length[u] = 0.0
                else:
                    self._singleton_code_length[u] = utils.log2(v, usage_sum)

        if len(self._edge_singleton_usage) != 0:
            for u, v in self._edge_singleton_usage.items():
                if utils.log2(v, usage_sum) == -0.0:
                    self._singleton_code_length[u] = 0.0
                else:
                    self._singleton_code_length[u] = utils.log2(v, usage_sum)

    def _compute_singleton_description_length(self):
        """ Compute the sum of each singleton pattern description length
        Returns
        -------
        float
        """
        desc = 0.0
        if len(self._vertex_singleton_usage.keys()) != 0:
            for key in self._vertex_singleton_usage.keys():
                desc += utils.encode_singleton(self._standard_table, 1, key)
                # desc += self._singleton_code_length[key]
                # port description, the singleton have one port and one node
                # Then we have 1 + len(nodes) = 2 and log2(binomial(1,1)), port code is 0
                desc += math.log2(2)

        if len(self._edge_singleton_usage.keys()) != 0:
            for key in self._edge_singleton_usage.keys():
                desc += utils.encode_singleton(self._standard_table, 2, key)
                # desc += self._singleton_code_length[key]
                # port description, the singleton have two ports and two nodes
                # Then we have 1 + len(nodes) = 3 and log2(binomial(2,2)), port code is {1:1, 2:1}
                # Then sum = 2
                desc += math.log2(3) + 2

        return desc

    def singleton_code_length(self):
        """ Provide the current singleton pattern code length
        Returns
        --------
        dict
        """
        return self._singleton_code_length

    def display_ct(self):
        """ Display the code table in a dataframe format
        Returns
        -------
        object
        """
        data = [row.display_row() for row in self._rows if row.pattern_usage() != 0]
        for u, v in self._singleton_code_length.items():
            data.append([u, v, '', '', '', ''])
        return pandas.DataFrame(numpy.array(data),
                                columns=['Pattern Structure', 'Pattern usage', 'Pattern code length',
                                         'Port count', 'Pattern port usage',
                                         'Pattern port code length'])

    def _order_rows(self, row: CodeTableRow):
        """ Provide the row elements to order rows
        Parameters
        ----------
        row
        Returns
        --------
        tuple
        """
        length = utils.get_total_label(row.pattern())
        support = utils.get_support(row.embeddings())
        return length, support, -row.arrival_number

    def is_node_marked(self, node_number, graph, cover_marker, label):
        """ Check if a particular node label  is already marked
        Parameters
        -----------
        node_number
        graph
        cover_marker
        label
        Returns
        ---------
        bool

        """
        if node_number in graph.nodes() and label in graph.nodes(data=True)[node_number]['label']:
            node = graph.nodes(data=True)[node_number]
            return 'cover_mark' in node and label in node['cover_mark'] and node['cover_mark'][label] == cover_marker
        else:
            raise ValueError(f"{node_number} should be a graph node and {label} a node label")

    def is_node_labels_marked(self, node_number, graph, cover_marker, labels):
        """ Check if one or more particular node label are marked
        Parameters
        -----------
        node_number
        graph
        cover_marker
        labels
        Returns
        --------
        bool"""
        if type(labels) is tuple:
            return not (False in [self.is_node_marked(node_number, graph, cover_marker, label) for label in labels])
        else:
            return self.is_node_marked(node_number, graph, cover_marker, labels)

    def mark_node(self, node_number, graph, cover_marker, label):
        """ Mark a node in the graph by the cover marker
        Parameters
        ---------
        node_number
        graph
        cover_marker
        label
        """

        if label is not None:
            # check if the pattern node are many labels
            if type(label) is tuple:  # if yes, mark each label if it is a graph node label
                for j in label:
                    if j in graph.nodes(data=True)[node_number]['label']:
                        if 'cover_mark' in graph.nodes(data=True)[node_number]:
                            graph.nodes(data=True)[node_number]['cover_mark'][j] = cover_marker
                        else:
                            graph.nodes(data=True)[node_number]['cover_mark'] = {j: cover_marker}

            else:  # if no, mark the unique label

                if 'cover_mark' in graph.nodes(data=True)[node_number]:
                    graph.nodes(data=True)[node_number]['cover_mark'][label] = cover_marker
                else:
                    graph.nodes(data=True)[node_number]['cover_mark'] = {label: cover_marker}
        else:
            raise ValueError("label shouldn't be empty or none")

    def _get_edge_index_from_label(self, values, label):
        """ Provide the good edge index, it's used in the multigraph treatment
        Parameters
        ---------
        values
        label
        Returns
        -------
        int
        """
        res = [i for i, j in enumerate(values, start=0) if
               'label' in j and j['label'] == label]
        if len(res) != 0:
            return res[0]
        else:
            return None

    def is_edge_marked(self, start, end, graph, cover_marker, label):
        """ Check if an edge is already marked
          Parameters
          -----------
          start
          end
          graph
          cover_marker
          label

          Returns
          ---------
          bool
          """
        if start in graph.nodes() and end in graph.nodes() and \
                graph[start][end] is not None:
            if type(graph) is nx.MultiDiGraph:
                if not (False in ['label' in v
                                  for v in graph.get_edge_data(start, end).values()]):
                    edge_index = self._get_edge_index_from_label(graph.get_edge_data(start, end).values(), label)
                    if edge_index is not None:
                        edge = graph[start][end][edge_index]
                        return 'cover_mark' in edge and label in edge['cover_mark'] and edge['cover_mark'][
                            label] == cover_marker
                    else:
                        return ValueError(f"{start} and {end} should be a graph node and {start}-{end} a graph edge."
                                          f"Also the edge should have {label}  as label")
                else:
                    raise ValueError(f"{start} and {end} should be a graph node and {start}-{end} a graph edge."
                                     f"Also the edge should have a label ")
            else:
                if 'label' in graph[start][end]:
                    edge = graph[start][end]
                    return 'cover_mark' in edge and label in edge['cover_mark'] and edge['cover_mark'][
                        label] == cover_marker
                else:
                    raise ValueError(f"{start} and {end} should be a graph node and {start}-{end} a graph edge."
                                     f"Also the edge should have a label ")

    def mark_edge(self, start, end, graph, cover_marker, label):
        """ Mark an edge in the graph by the cover marker
        Parameters
        ---------
        start
        end
        graph
        cover_marker
        label
        """
        if label is not None:
            if type(graph) is nx.MultiDiGraph:
                if not (False in ['label' in v
                                  for v in graph.get_edge_data(start, end).values()]):
                    edge_index = self._get_edge_index_from_label(graph.get_edge_data(start, end).values(), label)
                    if 'cover_mark' in graph[start][end][edge_index]:
                        graph[start][end][edge_index]['cover_mark'][label] = cover_marker
                    else:
                        graph[start][end][edge_index]['cover_mark'] = {label: cover_marker}
                else:
                    raise ValueError("label shouldn't be empty or none and must be an edge label ")

            else:
                if label in graph[start][end]['label']:
                    if 'cover_mark' in graph[start][end]:
                        graph[start][end]['cover_mark'][label] = cover_marker
                    else:
                        graph[start][end]['cover_mark'] = {label: cover_marker}
                else:
                    raise ValueError("label shouldn't be empty or none and must be an edge label ")

    def is_embedding_marked(self, embedding, pattern, graph, cover_marker):
        """ Check if an embedding is already marked
        Parameters
        ----------
        embedding
        pattern
        graph
        cover_marker

        Returns
        ---------
        bool
        """
        # keys = list(embedding.keys())
        # values = list(embedding.values())
        # i = 0
        res = []
        if len(utils.get_edge_in_embedding(embedding, pattern)):
            for edge in utils.get_edge_in_embedding(embedding, pattern):
                label = utils.get_edge_label(edge[0], edge[1], pattern)
                node1 = utils.get_key_from_value(embedding, edge[0])
                node2 = utils.get_key_from_value(embedding, edge[1])
                if (node1, node2) in list(graph.edges(node1)):
                    if type(label) is str:
                        if self.is_edge_marked(node1, node2, graph, cover_marker, label):
                            res.append(True)
                    elif type(label) is list:
                        for l in label:
                            if self.is_edge_marked(node1, node2, graph, cover_marker, l):
                                res.append(True)
                    else:
                        raise ValueError("All edges must have a label")
                elif (node2, node1) in list(graph.edges(node2)):
                    if type(label) is str:
                        if self.is_edge_marked(node1, node2, graph, cover_marker, label):
                            res.append(True)
                    elif type(label) is list:
                        for l in label:
                            if self.is_edge_marked(node1, node2, graph, cover_marker, l):
                                res.append(True)
                    else:
                        raise ValueError("All edges must have a label")
                else:
                    res.append(False)
            return True in res
        else:
            return ValueError("This embedding should have edges")

    def mark_embedding(self, embedding, graph, pattern, cover_marker):
        """ Mark an embedding of a pattern in the graph with a given cover_marker
        Parameters
        -----------
        embedding
        graph
        pattern
        cover_marker
        """
        if len(utils.get_edge_in_embedding(embedding, pattern)):
            for edge in utils.get_edge_in_embedding(embedding, pattern):
                label = utils.get_edge_label(edge[0], edge[1], pattern)
                node1 = utils.get_key_from_value(embedding, edge[0])
                node2 = utils.get_key_from_value(embedding, edge[1])
                if (node1, node2) in list(graph.edges(node1)):
                    if type(label) is str:
                        self.mark_edge(node1, node2, graph, cover_marker, label)
                    elif type(label) is list:
                        for l in label:
                            self.mark_edge(node1, node2, graph, cover_marker, l)
                    else:
                        raise ValueError("All edges must have label")
                elif (node2, node1) in list(graph.edges(node2)):
                    if type(label) is str:
                        self.mark_edge(node1, node2, graph, cover_marker, label)
                    elif type(label) is list:
                        for l in label:
                            self.mark_edge(node1, node2, graph, cover_marker, l)
                    else:
                        raise ValueError("All edges must have label")

                if 'label' in pattern.nodes(data=True)[edge[0]]:
                    self.mark_node(node1, graph, cover_marker, pattern.nodes(data=True)[edge[0]]['label'])

                if 'label' in pattern.nodes(data=True)[edge[1]]:
                    self.mark_node(node2, graph, cover_marker, pattern.nodes(data=True)[edge[1]]['label'])

        else:
            return ValueError("This embedding should have edges")

    def get_node_label_number(self, node_number, graph):
        """ compute the number of label for a given node_number in the graph
        Parameters
        ----------
        node_number
        graph
        Returns
        ---------
        int
        """
        if node_number in graph.nodes():
            if 'label' in graph.nodes(data=True)[node_number]:
                return len(graph.nodes(data=True)[node_number]['label'])
            else:
                return 0
        else:
            raise ValueError(f"{node_number} should be a graph node")

    def search_port(self, graph, embedding, cover_marker, pattern, port_usage):
        """ Search all node who are port for data or for pattern
        Parameters
        ----------
        graph
        embedding
        cover_marker
        port_usage
        pattern
        Returns
        -------
        list
        """
        res = []
        # Search pattern ports
        keys = list(embedding.keys())  # node number in graph
        values = list(embedding.values())  # node number in pattern
        i = 0
        while i <= len(keys) - 1:
            if not self.is_node_edges_marked(graph, keys[i], pattern, values[i]) \
                    or not self.is_node_all_labels_marked(keys[i], graph, cover_marker):
                res.append((keys[i], values[i]))
                if 'port' not in graph.nodes[keys[i]]:  # if the node is not already marked as port
                    graph.nodes[keys[i]]['port'] = True  # mark it

                if values[i] in port_usage:
                    port_usage[values[i]] = port_usage[values[i]] + 1
                else:
                    port_usage[values[i]] = 1
            i += 1

        return res

    def is_node_edges_marked(self, graph, graph_node, pattern, pattern_node):
        """ Check if the graph and the pattern corresponding node are the same edges number
        Parameters
        ----------
        graph
        graph_node
        pattern
        pattern_node

        Returns
        ---------
        bool
        """
        if graph_node in graph.nodes() and pattern_node in pattern.nodes():
            if graph.edges(graph_node) is not None and pattern.edges(pattern_node) is not None:
                if type(graph) is nx.Graph or type(graph) is nx.MultiGraph:
                    return len(graph.edges(graph_node)) == len(pattern.edges(pattern_node))
                else:
                    return len(graph.in_edges(graph_node)) + len(graph.out_edges(graph_node)) \
                           == len(pattern.in_edges(pattern_node)) + len(pattern.out_edges(pattern_node))
            else:
                return True
        else:
            raise ValueError(f"{graph_node} should be a nod of the graph and {pattern_node}, a node of pattern")

    def is_node_all_labels_marked(self, node_number, graph, cover_marker):
        """ Check if all node labels are marked by the cover_marker
        Parameters
        ----------
        node_number
        graph
        cover_marker

        Returns
        --------
        bool"""

        if node_number in graph.nodes():
            if 'label' in graph.nodes(data=True)[node_number]:
                response = []
                label = graph.nodes(data=True)[node_number]['label']
                if type(label) is str:
                    response.append(self.is_node_marked(node_number, graph, cover_marker, label))
                else:
                    for l in label:
                        response.append(self.is_node_marked(node_number, graph, cover_marker, l))

                return not (False in response)
            else:
                return True
        else:
            raise ValueError(f"{node_number} should be a graph node and must have a label ")

    def create_pattern_node(self, rewritten_graph, pattern_number, ports):
        """ Add a new node with its edges in the rewritten graph who represents a pattern embedding
           Parameters
           ----------
           rewritten_graph
           pattern_number : the pattern row number in the code table rows
           ports
           """
        if len(ports) == 1:
            last_node = len(rewritten_graph.nodes())
            rewritten_graph.add_node(last_node + 1, label=f"P{pattern_number}", is_Pattern=True)
            self.create_rewrite_edge(rewritten_graph, last_node + 1, ports[0][0], pattern_port=ports[0][1])
        elif len(ports) > 1:
            last_node = len(rewritten_graph.nodes())
            node = last_node + 1
            rewritten_graph.add_node(node, label=f"P{pattern_number}", is_Pattern=True)
            for p in ports:
                self.create_rewrite_edge(rewritten_graph, node, p[0], pattern_port=p[1])
        else:
            last_node = len(rewritten_graph.nodes())
            rewritten_graph.add_node(last_node + 1, label=f"P{pattern_number}", is_Pattern=True)

    def create_vertex_singleton_node(self, rewritten_graph, node_label, node_number):
        """ Add a new node with its edges in the rewritten graph who represents a vertex singleton embedding
           Parameters
           ----------
           rewritten_graph
           node_label
           node_number
           """
        last_node = len(rewritten_graph.nodes())
        rewritten_graph.add_node(last_node + 1, label=f"{node_label}", is_Pattern=True, is_singleton=True)
        self.create_rewrite_edge(rewritten_graph, last_node + 1, node_number, pattern_port=1)

    def create_edge_singleton_node(self, rewritten_graph, edge_label, first_node, second_node):
        """ Add a new node with its edges in the rewritten graph who represents an edge singleton embedding
        Parameters
        ----------
        rewritten_graph
        edge_label
        first_node
        second_node
        """
        last_node = len(rewritten_graph.nodes())
        rewritten_graph.add_node(last_node + 1, label=f"{edge_label}", is_Pattern=True, is_singleton=True)
        self.create_rewrite_edge(rewritten_graph, last_node + 1, first_node, pattern_port=1)
        self.create_rewrite_edge(rewritten_graph, last_node + 1, second_node, pattern_port=2)

    def create_rewrite_edge(self, rewritten_graph, rewritten_node, data_node, **kwargs):
        """ Add a new edge between a node and a port in the rewritten graph
        Parameters
        -----------
        rewritten_graph
        rewritten_node
        data_node : the port node in the initial data graph
        """
        if str(data_node) in dict(rewritten_graph.nodes(data='label')).values():
            port_node = utils.get_key_from_value(dict(rewritten_graph.nodes(data='label')), str(data_node))
            rewritten_graph.add_edge(rewritten_node, port_node, label=f"v{kwargs['pattern_port']}")
        else:
            last = len(rewritten_graph.nodes())
            rewritten_graph.add_node(last + 1, label=f"{data_node}")
            rewritten_graph.add_edge(rewritten_node, last + 1, label=f"v{kwargs['pattern_port']}")

    def row_cover(self, row: CodeTableRow, graph, cover_marker, rewritten_graph, row_number):
        """ Cover a code table row on the graph with a given marker
        Parameters
        ----------
        row
        graph
        cover_marker
        rewritten_graph
        row_number
        """
        cover_usage = 0
        port_usage = dict()
        if not utils.is_without_edge(row.pattern()):
            for embedding in row.embeddings():
                if not self.is_embedding_marked(embedding, row.pattern(), graph, cover_marker):
                    self.mark_embedding(embedding, graph, row.pattern(), cover_marker)
                    ports = self.search_port(graph, embedding, cover_marker, row.pattern(), port_usage)
                    row.add_used_embeddings((embedding, ports))
                    cover_usage = cover_usage + 1
                    # create the pattern in the rewritten graph
                    self.create_pattern_node(rewritten_graph, row_number, ports)
        else:
            for embedding in row.embeddings():
                keys = list(embedding.keys())
                values = list(embedding.values())
                i = 0
                while i <= len(keys) - 1:
                    label = row.pattern().nodes(data=True)[values[i]]['label']
                    # check if all pattern node labels are marked in the graph node
                    if not self.is_node_labels_marked(keys[i], graph, cover_marker, label):
                        ports = self.search_port(graph, embedding, cover_marker, row.pattern(), port_usage)
                        cover_usage = cover_usage + 1
                        # The embedding marking consist of the node label marking
                        self.mark_node(keys[i], graph, cover_marker, label)
                        row.add_used_embeddings((embedding, ports))
                        # create the pattern in the rewritten graph
                        self.create_pattern_node(rewritten_graph, row_number,
                                                 ports)  # with experimental rewritten graph implementation

                    i = i + 1

        row.set_pattern_port_usage(port_usage)
        row.set_pattern_usage(cover_usage)

    def singleton_cover(self, graph, cover_marker, rewritten_graph):
        """ Cover the code table with the singleton pattern
        Parameters
        -----------
        graph
        cover_marker
        rewritten_graph
        Returns
        --------
        tuple : who contains first dictionary for vertex usage and other for edge usage
        """
        edge_singleton_usage = defaultdict(int)
        vertex_singleton_usage = defaultdict(int)
        singleton_used_embeddings = defaultdict(lambda: [])
        # for edge label
        for edge in graph.edges(data=True):
            if not self.is_edge_marked(edge[0], edge[1], graph, cover_marker, edge[2]['label']):
                edge_singleton_usage[edge[2]['label']] += 1
                self.mark_edge(edge[0], edge[1], graph, cover_marker, edge[2]['label'])  # Mark the edge
                # Mark the edge nodes as data port
                graph.nodes[edge[0]]['port'] = True
                graph.nodes[edge[1]]['port'] = True
                singleton_used_embeddings[edge[2]['label']].append(({edge[0]: 1,
                                                                     edge[1]: 2},
                                                                    [(edge[0], 1), (edge[1], 2)]))
                # create the singleton embedding in the rewritten graph
                self.create_edge_singleton_node(rewritten_graph, edge[2]['label'], edge[0], edge[1])
        # for node label
        for node in graph.nodes(data=True):
            if 'label' in node[1]:
                if type(node[1]['label']) is not str:
                    for label in node[1]['label']:
                        if not self.is_node_marked(node[0], graph, cover_marker, label):
                            vertex_singleton_usage[label] += 1
                            self.mark_node(node[0], graph, cover_marker, label)
                            graph.nodes[node[0]]['port'] = True  # mark the node as port
                            singleton_used_embeddings[label].append(({node[0]: 1}, [node[0], 1]))  # store the
                            # singleton used embedding

                            # create the singleton embedding in the rewritten graph
                            self.create_vertex_singleton_node(rewritten_graph, label, node[0])
                else:
                    if not self.is_node_marked(node[0], graph, cover_marker, node[1]['label']):
                        vertex_singleton_usage[node[1]['label']] += 1
                        self.mark_node(node[0], graph, cover_marker, node[1]['label'])
                        graph.nodes[node[0]]['port'] = True  # mark as port
                        # create the singleton embedding in the rewritten graph
                        self.create_vertex_singleton_node(rewritten_graph, node[1]['label'], node[0])

        return vertex_singleton_usage, edge_singleton_usage, singleton_used_embeddings


    def data_is_multigraph(self):
        """ Provide if the data graph is a multigraph or not
        Returns
        -------
        bool
        """
        return type(self._data_graph) is nx.MultiDiGraph

    def __str__(self) -> str:
        msg = "\n Pattern |usage |code_length |port_count |port_usage |port_code \n"
        for row in self._rows:
            msg += str(row) + "\n"

        for u, v in self._singleton_code_length.items():
            msg += "{}|{}|{}".format(u, v, self._singleton_code_length[u]) + "\n"

        return msg



"""
Base IO code for all datasets
"""

import os


def get_data_home(data_home=None):
    """Return the path of the scikit-mine data home directory.

    This folder is used by some large dataset loaders to avoid downloading
    data several times.

    By default, ``data_home`` is **./scikit_mine_data/**

    Alternatively, it can be set by the **SCIKIT_MINE_DATA** environment
    variable or programmatically by giving an explicit folder path.

    Parameters
    ----------
    data_home : str | None
        The path to scikit-mine data dir.
    """
    import skmine
    sk_path = skmine.__file__  # example sk_path = '/home/username/codes/sk_root/skmine/__init__.py'
    sk_root = os.path.dirname(os.path.dirname(sk_path))  # sk_root = '/home/username/codes/sk_root

    if data_home is None:
        data_home = os.environ.get("SCIKIT_MINE_DATA", os.path.join(sk_root, "scikit_mine_data"))
    data_home = os.path.expanduser(data_home)
    if not os.path.exists(data_home):
        os.makedirs(data_home)
    return data_home



"""
The :mod:`skmine.graph` module includes methods which consider recurrent subgraphs called pattern in a data graph .
A graph is a structure that represents data as a set of entities called "nodes", connected by relationships
called "edges". In a labeled graph, nodes and edges can have associated labels that allow to represent attributes of entities and relationships.
"""


"""
utils functions
"""

import numbers
from itertools import count

import numpy as np
import pandas as pd
import scipy.sparse
from numpy.core.numeric import normalize_axis_tuple
from numpy.core.overrides import array_function_dispatch
from numpy.lib.stride_tricks import as_strided
from sortedcontainers import SortedList


def _check_random_state(seed):
    if seed is None or seed is np.random:
        return np.random.mtrand._rand
    if isinstance(seed, numbers.Integral):
        return np.random.RandomState(seed)
    if isinstance(seed, np.random.RandomState):
        return seed
    raise ValueError(
        f"{seed} cannot be used to seed a numpy.random.RandomState instance"
    )


def _check_min_supp(min_supp, accept_absolute=True):
    if isinstance(min_supp, int):
        if not accept_absolute:
            raise ValueError(
                "Absolute support is prohibited, please provide a float value between 0 and 1"
            )
        if min_supp < 1:
            raise ValueError("Minimum support must be strictly positive")
    elif isinstance(min_supp, float):
        if min_supp < 0 or min_supp > 1:
            raise ValueError("Minimum support must be between 0 and 1")
    else:
        raise TypeError("Mimimum support must be of type int or float")
    return min_supp


def _check_growth_rate(gr):
    if not gr > 1:
        raise ValueError("growth ratio should be greater than 1")
    return gr


def filter_maximal(itemsets):
    """filter maximal itemsets from a set of itemsets

    Parameters
    ----------
    itemsets: Iterator[list]
        a list of itemsets

    Returns
    -------
    SortedList
    """
    maximals = SortedList(key=len)
    itemsets = sorted(itemsets, key=len, reverse=True)
    for itemset_set in itemsets:
        gts = maximals.irange(itemset_set)
        # # is there a superset amongst bigger itemsets ?
        if not any(map(lambda e: e > itemset_set, gts)):
            maximals.add(itemset_set)  # O(log(len(maximals)))

    return maximals


def filter_minimal(itemsets):
    """filter minimal itemsets from a set of itemsets

    Parameters
    ----------
    itemsets: Iterator[frozenset]
        a set of itemsets

    Returns
    -------
    SortedList
    """
    minimals = SortedList(key=len)
    itemsets = sorted(itemsets, key=len)
    for iset in itemsets:
        lts = minimals.irange(None, iset)
        # is there a subset amongst the smaller itemsets ?
        if not any(map(lambda e: e < iset, lts)):
            minimals.add(iset)

    return minimals


def supervised_to_unsupervised(D, y):
    """
    for sklearn compatibility, eg. sklearn.multiclass.OneVSRest

    Parameters
    ----------
    D: pd.DataFrame
        input transactional dataset

    y: np.ndarray of shape (n_samples,)
        corresponding labels
    """
    mask = np.where(y.reshape(-1))[0]
    D = D.iloc[mask]

    return D


def _check_D_sklearn(D):
    if object in D.dtypes.values:  # SKLEARN : check_dtype_object
        raise TypeError("argument must be a string or a number")

    if D.shape[1] == 0:  # SKLEARN : check_empty_data_messages
        raise ValueError("Empty data")

    pd.options.mode.use_inf_as_na = True
    if D.isnull().values.any():
        raise ValueError("esimator does not check for NaN and inf")
    pd.options.mode.use_inf_as_na = False


def _check_D(D):
    if isinstance(D, pd.DataFrame):
        D = D.reset_index(drop=True)  # positional indexing
    elif isinstance(D, np.ndarray):
        D = pd.DataFrame(D)
    elif scipy.sparse.issparse(D):
        D = pd.DataFrame.sparse.from_spmatrix(D)
    else:
        raise TypeError("D should be an instance of np.ndarray or pd.DataFrame")

    _check_D_sklearn(D)

    return D


def _check_y(y):
    if not isinstance(y, (pd.Series, np.ndarray)):
        raise TypeError("y should be an instance of np.ndarray or pd.Series")

    # TODO : pd.Categorical
    return y


def _check_D_y(D, y=None):
    D = _check_D(D)
    if y is not None:
        y = _check_y(y)
    return D, y


def intersect2d(ar1, ar2, return_indices=True):
    """
    Find the intersection of two 2 dimnesional arrays

    Return the sorted, unique rows that are both of the input arrays

    Parameters
    ----------
    x, y: array_like
        Input arrays
    return_indices: bool
        If True, the indices wich correspond to the intersection of the two array
        are returned. The first instance of a value is used if there are multiple.
        Default is False

    Returns
    -------
    intersect2d: ndarray
        Sorted 2D array of common and unique rows
    comm1: ndarray
        The indices of the first occurences of the common rows in `ar1`.
        Only provided if `return_indices` is True
    comm2: ndarray
        The indices of the first occurences of the common rows in `ar2`.
        Only provided if `return_indices` is True
    """
    ar1, ar2 = np.asanyarray(ar1), np.asanyarray(ar2)
    assert ar1.ndim == ar2.ndim == 2
    _, x_ind, y_ind = np.intersect1d(ar1[:, 0], ar2[:, 0], return_indices=True)
    x, y = ar1[x_ind], ar2[y_ind]
    xy_where = np.argwhere((x == y).all(axis=1)).reshape(-1)
    x_ind = x_ind[xy_where]
    if return_indices:
        y_ind = y_ind[xy_where]
        return ar1[x_ind], x_ind, y_ind
    return ar1[x_ind]


def _sliding_window_view_dispatcher(
    x, window_shape, axis=None, *, subok=None, writeable=None
):
    return (x,)


@array_function_dispatch(_sliding_window_view_dispatcher)
def sliding_window_view(x, window_shape, axis=None, *, subok=False, writeable=False):
    """
    COPIED from https://github.com/numpy/numpy/blob/v1.20.0/numpy/lib/stride_tricks.py#L122-L336

    Create a sliding window view into the array with the given window shape.
    Also known as rolling or moving window, the window slides across all
    dimensions of the array and extracts subsets of the array at all window
    positions.

    Parameters
    ----------
    x : array_like
        Array to create the sliding window view from.
    window_shape : int or tuple of int
        Size of window over each axis that takes part in the sliding window.
        If `axis` is not present, must have same length as the number of input
        array dimensions. Single integers `i` are treated as if they were the
        tuple `(i,)`.
    axis : int or tuple of int, optional
        Axis or axes along which the sliding window is applied.
        By default, the sliding window is applied to all axes and
        `window_shape[i]` will refer to axis `i` of `x`.
        If `axis` is given as a `tuple of int`, `window_shape[i]` will refer to
        the axis `axis[i]` of `x`.
        Single integers `i` are treated as if they were the tuple `(i,)`.
    subok : bool, optional
        If True, sub-classes will be passed-through, otherwise the returned
        array will be forced to be a base-class array (default).
    writeable : bool, optional
        When true, allow writing to the returned view. The default is false,
        as this should be used with caution: the returned view contains the
        same memory location multiple times, so writing to one location will
        cause others to change.
    Returns
    -------
    view : ndarray
        Sliding window view of the array. The sliding window dimensions are
        inserted at the end, and the original dimensions are trimmed as
        required by the size of the sliding window.
        That is, ``view.shape = x_shape_trimmed + window_shape``, where
        ``x_shape_trimmed`` is ``x.shape`` with every entry reduced by one less
        than the corresponding window size.
    See Also
    --------
    lib.stride_tricks.as_strided: A lower-level and less safe routine for
        creating arbitrary views from custom shape and strides.
    broadcast_to: broadcast an array to a given shape.
    Notes
    -----
    For many applications using a sliding window view can be convenient, but
    potentially very slow. Often specialized solutions exist, for example:
    - `scipy.signal.fftconvolve`
    - filtering functions in `scipy.ndimage`
    - moving window functions provided by
      `bottleneck <https://github.com/pydata/bottleneck>`_.
    As a rough estimate, a sliding window approach with an input size of `N`
    and a window size of `W` will scale as `O(N*W)` where frequently a special
    algorithm can achieve `O(N)`. That means that the sliding window variant
    for a window size of 100 can be a 100 times slower than a more specialized
    version.
    Nevertheless, for small window sizes, when no custom algorithm exists, or
    as a prototyping and developing tool, this function can be a good solution.
    Examples
    --------

    >>> x = np.arange(6)
    >>> x.shape
    (6,)
    >>> v = sliding_window_view(x, 3)
    >>> v.shape
    (4, 3)
    >>> v
    array([[0, 1, 2],
           [1, 2, 3],
           [2, 3, 4],
           [3, 4, 5]])
    >>> moving_average = v.mean(axis=-1)
    >>> moving_average
    array([1., 2., 3., 4.])
    """
    window_shape = tuple(window_shape) if np.iterable(window_shape) else (window_shape,)
    # first convert input to array, possibly keeping subclass
    x = np.array(x, copy=False, subok=subok)

    window_shape_array = np.array(window_shape)
    if np.any(window_shape_array < 0):
        raise ValueError("`window_shape` cannot contain negative values")

    if axis is None:
        axis = tuple(range(x.ndim))
        if len(window_shape) != len(axis):
            raise ValueError(
                f"Since axis is `None`, must provide "
                f"window_shape for all dimensions of `x`; "
                f"got {len(window_shape)} window_shape elements "
                f"and `x.ndim` is {x.ndim}."
            )
    else:
        axis = normalize_axis_tuple(axis, x.ndim, allow_duplicate=True)
        if len(window_shape) != len(axis):
            raise ValueError(
                f"Must provide matching length window_shape and "
                f"axis; got {len(window_shape)} window_shape "
                f"elements and {len(axis)} axes elements."
            )

    out_strides = x.strides + tuple(x.strides[ax] for ax in axis)

    # note: same axis can be windowed repeatedly
    x_shape_trimmed = list(x.shape)
    for ax, dim in zip(axis, window_shape):
        if x_shape_trimmed[ax] < dim:
            raise ValueError("window shape cannot be larger than input array shape")
        x_shape_trimmed[ax] -= dim - 1
    out_shape = tuple(x_shape_trimmed) + window_shape
    return as_strided(
        x, strides=out_strides, shape=out_shape, subok=subok, writeable=writeable
    )


def bron_kerbosch(candidates: dict, clique=None, excluded=None, depth=0):
    """
    Bron-Kerbosch algorithm, from https://en.wikipedia.org/wiki/Bron%E2%80%93Kerbosch_algorithm

    Parameters
    ----------
    candidates: dict[object, list[object]]
        a mapping from each node to its existing neighbours
    """
    if not candidates and not excluded and clique is not None and len(clique) > 2:
        yield [_ for _ in clique]
        return

    # pass None as default arg instead of dict()
    # fix collisions in pytest, really obscure
    clique = clique or dict()
    excluded = excluded or dict()

    if depth > 20:
        return

    for node, neighbours in list(candidates.items()):
        new_clique = {**{node: neighbours}, **clique}
        new_candidates = {k: v for k, v in candidates.items() if k in neighbours}
        new_excluded = {k: v for k, v in excluded.items() if k in neighbours}

        yield from bron_kerbosch(new_candidates, new_clique, new_excluded)

        del candidates[node]
        excluded[node] = neighbours




"""Base classes for all miners."""
# pylint: disable= unused-argument

import inspect
from abc import ABC, abstractmethod

import numpy as np
import pandas as pd


def _get_tags(self):
    return {
        "non_deterministic": False,
        "requires_positive_X": False,
        "requires_positive_y": False,
        "X_types": ["2darray"],
        "poor_score": False,
        "no_validation": True,
        "multioutput": False,
        "allow_nan": False,
        "stateless": False,
        "multilabel": False,
        "_skip_test": False,
        "_xfail_checks": False,
        "multioutput_only": False,
        "binary_only": False,
        "requires_fit": True,
        "preserves_dtype": [np.float64],
        "requires_y": False,
        "pairwise": False,
    }


class BaseMiner(ABC):
    """Base class for all miners in scikit-mine."""

    @abstractmethod
    def fit(self, D, y):
        """Fit method to be implemented."""
        return self

    @abstractmethod
    def discover(self, *args, **kwargs):
        """discover method to be implemented."""
        return pd.Series()

    _get_tags = _get_tags

    @classmethod
    def _get_param_names(cls):
        """Get parameter names for the estimator"""
        # fetch the constructor or the original constructor before
        # deprecation wrapping if any
        init = getattr(cls.__init__, "deprecated_original", cls.__init__)
        if init is object.__init__:
            # No explicit constructor to introspect
            return []

        # introspect the constructor arguments to find the model parameters
        # to represent
        init_signature = inspect.signature(init)
        # Consider the constructor parameters excluding 'self'
        parameters = [
            p
            for p in init_signature.parameters.values()
            if p.name != "self" and p.kind != p.VAR_KEYWORD
        ]
        # Extract and sort argument names excluding 'self'
        return sorted([p.name for p in parameters])

    def get_params(self, deep=False):
        """
        Get parameters for this estimator.

        Returns
        -------
        params : mapping of string to any
            Parameter names mapped to their values.
        """
        out = dict()
        for key in self._get_param_names():
            value = getattr(self, key)
            out[key] = value
        return out

    def set_params(self, **params):
        """
        Set the parameters of this estimator.
        The method works on simple estimators as well as on nested objects
        (such as pipelines). The latter have parameters of the form
        ``<component>__<parameter>`` so that it's possible to update each
        component of a nested object.

        Parameters
        ----------
        **params : dict
            Estimator parameters.
        Returns
        -------
        self : object
            Estimator instance.
        """
        # Simple optimization to gain speed (inspect is slow)
        if not params:
            return self

        valid_params = self.get_params()

        for key, value in params.items():
            if key not in valid_params:
                raise ValueError(
                    "Invalid parameter %s for estimator %s. "
                    "Check the list of available parameters "
                    "with `estimator.get_params().keys()`." % (key, self)
                )

            setattr(self, key, value)
            valid_params[key] = value

        return self


class DiscovererMixin:
    """Mixin for all pattern discovery models in scikit-mine"""

    def fit_discover(self, D, y=None, **kwargs):
        """
        Fit to data, the extract patterns

        Parameters
        ----------
        D: {array-like, sparse matrix, dataframe} of shape (n_samples, n_features)

        Returns
        -------
        pd.Series
            patterns discovered by a mining algorithm
        """
        if y is None:
            return self.fit(D).discover(**kwargs)
        return self.fit(D, y=y).discover(**kwargs)


class TransformerMixin:
    """Base Mixin for transformers in scikit-mine"""

    def fit_transform(self, X, y=None):
        "fit on X and y, then transform X"
        return self.fit(X, y).transform(X)

    decision_function = None


class MDLOptimizer(ABC):
    """
    Base interface for all models applying the `Minimum Description Length principle
    <https://en.wikipedia.org/wiki/Minimum_description_length>`_.
    """

    @abstractmethod
    def generate_candidates(self, *args, **kwargs):
        """
        Generate new candidates, to be sent for later evaluation.

        Calling this function is equivalent to sending a new message given an encoding scheme,
        while calling ``.evaluate`` is equivalent to receiving this message, and evaluating the gain
        of information it provides.

        Returns
        -------
        object or Iterable[object]
            A set of new candidates
        """
        return list()

    @abstractmethod
    def evaluate(self, candidate, *args, **kwargs):
        """
        Evaluate the gain, i.e the gain of information when accepting the candidate.

        Parameters
        ----------
        candidate: object
            A candidate to evaluate

        Returns
        -------
        tuple (data_size, model_size, ...)
            Should return a tuple, with first two values corresponding to new data size
            and model size in the case of accepting the candidate.

            Data size and model size should be returned separately as we encourage
            usage of `(two-part) crude MDL
            <https://en.wikipedia.org/wiki/Minimum_description_length#Two-Part_Codes>`_.
        """
        return (
            0,
            0,
        )

    def _repr_html_(self):
        s = self.discover()  # call discover with default parameters
        df = s.to_frame(name="usage")
        if not df.empty:
            return df._repr_html_()  # pylint: disable=protected-access
        return repr(self)


class InteractiveMiner(ABC):
    """Base class for interactive mining

    Interactive miners should allow us to
    1. ingest some input data, by calling `prefit`
    2. generate candidates
    3. loop over generated candidate, and call `update` with this candidate as argument,
    depending on some external input (like a positive answer from a user in CLI mode)
    """

    @abstractmethod
    def prefit(self, D):
        """ingest data `D` and track basic informations to be used later"""
        return self

    @abstractmethod
    def update(self, *args, **kwargs):
        """inplace edition of underlying datastructures"""
        return None



"""
Multi-Interval Discretization of Continuous-Valued Attributes for Classification Learning
"""

# Author: Rémi Adon <remi.adon@gmail.com>
#
# License: BSD 3 clause

import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from scipy.stats import entropy

from skmine.base import BaseMiner, MDLOptimizer
from skmine.utils import _check_random_state


def get_entropy_nb_ones(y):
    """
    for a given vector y, returns the entropy (base 2) and the
    number of non zeros values in the one-hot-encoded counting vector
    associated with y
    """
    ohe = np.bincount(y) / len(y)  # counts, one hot encoded
    return entropy(ohe, base=2), np.sum(ohe != 0)


def generate_cut_point(y, start, end):
    """
    Generate a cut point given a label vector ``y``, a start position
    ``start`` and a final position ``end``

    Starts with an infinite entropy, then iteratively moves a index inside
    ``y``, pre-evaluates entropy on both sides of this index.

    It returns the index with the minimum entropy.
    """
    length = end - start
    ent = np.inf
    k = -1

    for idx in range(start + 1, end):
        if y[idx - 1] == y[idx]:
            continue

        first_half_ent = get_entropy_nb_ones(y[start:idx])[0]
        first_half_ent *= (idx - start) / length

        second_half_ent = get_entropy_nb_ones(y[idx:end])[0]
        second_half_ent *= (end - idx) / length

        new_ent = first_half_ent + second_half_ent

        if new_ent < ent:
            ent = new_ent
            k = idx

    return k


class MDLPVectDiscretizer(MDLOptimizer):
    """
    Basic block for the implementation of
    "Multi-Interval Discretization of Continuous-Valued Attributes for Classification Learning".

    This class operates at a column level, i.e it finds the best cut points for a given feature
    to fit a the corresponding labels
    """

    def __init__(self, min_depth=0):
        self.min_depth_ = min_depth
        self.entropy_ = np.inf
        self.cut_points_ = np.array([])

    generate_candidates = generate_cut_point

    def evaluate(self, y, start, end, cut_point):
        """
        Evaluate vector y of size ``end`` - ``start``,
        given a ``cutpoint``
        """
        entropy1, k1 = get_entropy_nb_ones(y[start:cut_point])
        entropy2, k2 = get_entropy_nb_ones(y[cut_point:end])
        whole_entropy, k0 = get_entropy_nb_ones(y[start:end])

        N = end - start

        part1 = 1 / N * ((cut_point - start) * entropy1 + (end - cut_point) * entropy2)
        delta = np.log2(pow(3, k0) - 2) - (
            k0 * whole_entropy - k1 * entropy1 - k2 * entropy2
        )

        gain = whole_entropy - part1

        return gain > 1 / N * (np.log2(N - 1) + delta)

    def fit(self, X, y):
        """
        fit discretizer on a single feature

        Parameters
        ----------
        X : numpy.ndarray of shape (n_samples,)
            Input vector to fit the discretizer on

        y : numpy.ndarray of shape (n_samples,)
            Labels to fit the discretizer on
        """
        assert len(X.shape) == 1
        assert np.issubdtype(y.dtype, np.integer)

        order = np.argsort(X)
        X = X[order]
        y = y[order]

        cut_points = set()
        search_intervals = [(0, len(X), 0)]
        while search_intervals:
            level = search_intervals.pop(-1)  # pop back
            start, end, depth = level

            k = generate_cut_point(y, start, end)

            is_better = self.evaluate(y, start, end, k)
            if k == -1 or not is_better:
                front = -np.inf if start == 0 else (X[start - 1] + X[start]) / 2
                back = np.inf if end == len(X) else (X[end - 1] + X[end]) / 2

                if front != -np.inf:
                    cut_points.add(front)
                if back != np.inf:
                    cut_points.add(back)
            else:
                search_intervals.append((start, k, depth + 1))
                search_intervals.append((k, end, depth + 1))

        self.cut_points_ = np.array(list(cut_points))
        return self


class MDLPDiscretizer(BaseMiner):
    """
    Implementation of "Multi-Interval Discretization of Continuous-Valued Attributes
    for Classification Learning".

    Given class labels ``y``, MDLPDIscretizer discretizes continuous variables from
    ``X`` by minimizing the entropy in each interval.

    Parameters
    ----------
    random_state : int, RandomState instance, default=None
        random state to use to shuffle the data. Can affect the outcome, leading to
        slightly different cut points if a variable contains samples with the same value
        but different labels.

    Attributes
    ----------
    cut_points_: dict
        A mapping between columns and their respective cut points.
        If fitted on a pandas DataFrame, keys will be the DataFrame column names.


    References
    ----------
    Usama M. Fayyad, Keki B. Irani
    "Multi-Interval Discretization of Continuous-Valued Attributes for Classification Learning",
    1993

    Examples
    --------

    >>> from skmine.preprocessing import MDLPDiscretizer
    >>> from sklearn.datasets import load_iris  # doctest: +SKIP
    >>> iris = load_iris()                      # doctest: +SKIP
    >>> X, y = iris.data, iris.target           # doctest: +SKIP
    >>> disc = MDLPDiscretizer()                # doctest: +SKIP
    >>> disc.fit(X, y)                          # doctest: +SKIP
    >>> disc.cut_points_                        # doctest: +SKIP
    {0: array([5.5, 6.2]), 1: array([2.9, 3.3]), 2: array([2.45, 4.9 ]), 3: array([0.8, 1.7])}

    """

    def __init__(self, random_state=None, n_jobs=1):
        self.cut_points_ = dict()
        self.random_state = random_state
        self.n_jobs = n_jobs
        self.discretizers_ = []

    def fit(self, X, y):
        """fit the MLDP discretizer on an input matrix ``X``, given a label vector ``y``.

        Parameters
        ----------
        X: np.ndarray or pd.DataFrame of shape (n_samples, n_features)
            The input matrix containing features. A set of cut points
            will be affected to each feature

        y : np.ndarray of pd.Series of shape(n_samples,)
            The label vector used to discretize ``X``
        """
        assert y is not None and np.issubdtype(y.dtype, np.integer)
        random_state = _check_random_state(self.random_state)
        permutation = random_state.permutation(len(X))
        _X = X.values if isinstance(X, pd.DataFrame) else X
        _y = y.values if isinstance(y, pd.Series) else y
        _X = _X[permutation]
        _y = _y[permutation]

        n_cols = _X.shape[1]

        discs = Parallel(n_jobs=self.n_jobs, prefer="processes")(
            delayed(MDLPVectDiscretizer().fit)(_X[:, idx], _y) for idx in range(n_cols)
        )

        self.discretizers_ = discs
        cut_points = [d.cut_points_ for d in discs]

        if isinstance(X, pd.DataFrame):
            self.cut_points_ = dict(zip(X.columns, cut_points))
        else:
            self.cut_points_ = dict(enumerate(cut_points))

        return self

    def discover(self):
        """user-friendly view on cut points"""
        return pd.Series(self.cut_points_, dtype=object)

    def transform(self, X, y=None):  # pylint: disable=unused-argument
        """Discretizes the input matrix X

        This applies the cutpoints their respective columns
        """
        if isinstance(X, pd.DataFrame) and not set(self.cut_points_) == set(X.columns):
            raise ValueError(f"X columns should be {self.cut_points_.keys()}")
        _X = X.values if isinstance(X, pd.DataFrame) else X

        vects = Parallel(n_jobs=self.n_jobs, prefer="threads")(
            delayed(np.searchsorted)(cut_points, _X[:, idx])
            for idx, cut_points in enumerate(self.cut_points_.values())
        )

        vects = [v.reshape(len(v), 1) for v in vects]
        _X = np.concatenate(vects, axis=1)

        if isinstance(X, pd.DataFrame):
            _X = pd.DataFrame(_X, columns=X.columns)

        return _X

    def fit_transform(self, X, y=None):
        "fit on X and y, then transform X"
        return self.fit(X, y).transform(X)



import networkx as nx


class Candidate:
    """
        It represents candidate whose format is <P1,P2,{(v1,v1)..}> in the algorithm,
        here, P1 and P2 are patterns and the last parameters, a port list

        Parameters
        -----------
        first_pattern_label : str
            It's the candidate first pattern name
        second_pattern_label : str
            It's the candidate second pattern name
        port: list
            It's the candidate port list
    """

    def __init__(self, first_pattern_label, second_pattern_label, port):
        self.first_pattern_label = first_pattern_label
        self.second_pattern_label = second_pattern_label
        self.first_pattern = None
        self.second_pattern = None
        self.port = port  # list of the candidate port association
        self.data_port = set()  # candidate port number in the rewritten graph
        self.usage = 0  # estimated usage
        self.exclusive_port_number = 0
        self._final_pattern = None  # merge pattern
        self.code_length = 0.0  # merge pattern description length

    def set_usage(self, usage):
        """ Set candidate estimated usage
        Parameters
        ----------
        usage
        """
        self.usage = usage

    def final_pattern(self):
        """ Provide the candidate merge pattern
        Returns
        --------
        object
        """
        if self._final_pattern is None:
            self._final_pattern = self.merge_candidate()
            return self._final_pattern
        else:
            return self._final_pattern

    def merge_candidate(self):
        """ Merge a candidate pattern
        Returns
        --------
        Graph
        """
        if type(self.first_pattern) is nx.MultiDiGraph:
            graph = nx.MultiDiGraph()
        else:
            graph = nx.DiGraph()
        ports = []
        for p in self.port:
            port = (int(p[0].split('v')[1]), int(p[1].split('v')[1]))
            ports.append(port)
        # Create first pattern node and edges
        self.create_candidate_first_pattern(self.first_pattern, graph)
        # Create second pattern node and edges
        self.create_candidate_second_pattern(self.second_pattern, graph, ports)
        return graph

    def create_candidate_first_pattern(self, pattern, graph):
        """ Create a given candidate first pattern nodes and edges
        Parameters
        ---------
        pattern
        graph
        """
        # create the pattern nodes with their labels in the given graph
        for node in pattern.nodes(data=True):
            if 'label' in node[1]:
                graph.add_node(node[0], label=node[1]['label'])
            else:
                graph.add_node(node[0])
        # create the pattern edges with their labels  in the given graph
        for edge in pattern.edges(data=True):
            graph.add_edge(edge[0], edge[1], label=edge[2]['label'])

    def create_candidate_second_pattern(self, pattern, graph, ports):
        """ Create a given candidate second pattern nodes and edges,
            and connected it with the existent graph elements by the port
            Parameters
            ---------
            pattern
            graph
            ports
        """
        mapping = dict()  # Mapping between pattern nodes and graph nodes number
        second_port = [p[1] for p in ports]  # pattern ports number
        first_port = [p[0] for p in ports]  # graph ports number

        # create the non-port pattern nodes with their labels in the graph
        # for the pattern port node add only the labels if it's necessary
        for node in pattern.nodes(data=True):
            if node[0] not in second_port:
                mapping[node[0]] = len(graph.nodes()) + 1
                if 'label' in node[1]:
                    graph.add_node(len(graph.nodes()) + 1, label=node[1]['label'])
                else:
                    graph.add_node(len(graph.nodes()) + 1)
            else:
                index = second_port.index(node[0])
                if 'label' in node[1]:
                    if 'label' in graph.nodes[first_port[index]]:
                        new_label = self._get_new_label(node[1]['label'], graph.nodes[first_port[index]]['label'])
                        graph.nodes[first_port[index]]['label'] = new_label
                    else:
                        graph.nodes[first_port[index]]['label'] = node[1]['label']

        # Set the mapping for the pattern port node
        for p in ports:
            mapping[p[1]] = p[0]

        # Create pattern edges with their labels
        for edge in pattern.edges(data=True):
            if edge[0] in second_port:
                port = first_port[second_port.index(edge[0])]
                graph.add_edge(port, mapping[edge[1]], label=edge[2]['label'])
            elif edge[1] in second_port:
                port = first_port[second_port.index(edge[1])]
                graph.add_edge(mapping[edge[0]], port, label=edge[2]['label'])
            elif edge[0] == edge[1] and edge[0] in port[1]:
                port = first_port[second_port.index(edge[1])]
                graph.add_edge(port[0], port, label=edge[2]['label'])
            else:
                graph.add_edge(mapping[edge[0]], mapping[edge[1]], label=edge[2]['label'])

    def _get_new_label(self, first_label, second_label):
        """ Associate two given list of label
        Parameters
        -----------
        first_label
        second_label
        Returns
        -------
        list
        """
        if type(second_label) is str:
            if type(first_label) is str:
                if first_label != second_label:
                    return first_label, second_label
                else:
                    return second_label
            else:
                label = list()
                label.append(second_label)
                for l in first_label:
                    if l not in label:
                        label.append(l)
                return label
        else:
            label = list(second_label)
            if type(first_label) is str:
                if first_label not in second_label:
                    label.append(first_label)
            else:
                for l in first_label:
                    if l not in label:
                        label.append(l)
            return label

    def _is_ports_equals(self, ports):
        """ Check if the candidate ports list are similar to a given ports list
        Parameters
        ----------
        ports
        Returns
        -------
        bool"""
        if len(ports) != len(self.port):
            return False
        else:
            return not (False in [p in ports for p in self.port])

    def compute_description_length(self, label_codes):
        """ Compute description length from the label codes to the candidate merge pattern
        Parameters
        ----------
        label_codes
        """
        if self.final_pattern is not None and label_codes is not None:
            self.code_length = label_codes.encode(self.final_pattern())
        else:
            raise ValueError("You should create the final pattern and set the label codes before computing")

    def inverse(self):
        """ Provide the candidate inverse
        Returns
        -------
        Candidate
        """
        ports = []
        for p in self.port:
            ports.append((p[1], p[0]))
        c = self.first_pattern_label
        self.first_pattern_label = self.second_pattern_label
        self.second_pattern_label = c
        self.port = ports
        del c

        return self

    def __str__(self) -> str:
        return "<{},{},{}>".format(self.first_pattern_label, self.second_pattern_label, self.port)

    def __eq__(self, o: object) -> bool:
        return o.first_pattern_label == self.first_pattern_label \
               and o.second_pattern_label == self.second_pattern_label \
               and self._is_ports_equals(o.port)



"""
utils methods for skmine.datasets
"""

from itertools import chain


def describe(D):
    """Give some high level properties on transactions

    ========================    ===============
    Number of items             int
    Number of transactions      int
    Average transaction size    float
    Density                     float in [0, 1]
    ========================    ===============

    Parameters
    ----------
    D: pd.Series
        A transactional dataset

    Notes
    -----
        .. math:: density = { avg\_transaction\_size \over n\_items }

    Example
    -------
    >>> from skmine.datasets.fimi import fetch_connect
    >>> from skmine.datasets.utils import describe
    >>> describe(fetch_connect())  # doctest: +SKIP
    {'n_items': 75, 'avg_transaction_size': 37.0, 'n_transactions': 3196, 'density': 0.4933}
    """
    avg_transaction_size = D.map(len).mean()
    n_transactions = D.shape[0]
    n_items = len(set(chain(*D)))
    return dict(
        n_items=n_items,
        avg_transaction_size=avg_transaction_size,
        n_transactions=n_transactions,
        density=avg_transaction_size / n_items,
    )


def describe_logs(D):
    """Give some high level properties on logs

    ==============================      =====
    Number of events                    int
    Average delta per event             float
    Average nb of points per event      float
    ==============================      =====

    Parameters
    ----------
    D: pd.Series
        A dataset containing logs

    Example
    -------
    >>> from skmine.datasets.periodic import fetch_health_app
    >>> from skmine.datasets.utils import describe_logs
    >>> describe(fetch_health_app()) # doctest: +SKIP
    {'n_events': 20,
    'avg_delta_per_event': Timedelta('0 days 00:53:24.984000'),
    'avg_nb_points_per_event': 100.0}
    """
    gb = D.groupby(D.values)
    a = gb.apply(lambda df: (df.index.max() - df.index.min(), len(df)))
    avg_nb_points = a.str[1].mean()

    return dict(
        n_events=len(gb),
        avg_delta_per_event=a.str[0].mean(),
        avg_nb_points_per_event=avg_nb_points,
    )



"""
MBD-LLBorder
"""
from functools import partial
from itertools import combinations

import numpy as np
import pandas as pd
from joblib import Parallel, delayed

from ..base import BaseMiner, DiscovererMixin
from ..itemsets.lcm import LCMMax
from ..utils import _check_growth_rate, _check_min_supp, filter_maximal, filter_minimal


def border_diff(U, S):
    """
    Given a pair of borders <{∅}, {U}> and <{∅}, {S}>,
    ``border_diff`` derives another border <L2, {U}>
    such that [L2, {U}] = [{∅}, {U}] - [{∅}, {S}]

    Parameters
    ----------
    U : set
        non empty part from the border to differentiate

    S : list of set
        non-empty part of a border.
        Noted as ``R1`` in the original paper

    References
    ----------
    .. [1]
        Dong, Li
        Efficient Mining of Emerging Patterns Discovering

    Notes
    -----
    See ``BORDER-DIFF`` in section 4.1
    """
    # assert len(R1) < len(U)  # assert we iterate on the smallest ensemble
    L2 = [{x} for x in U - S[0]]
    for i in range(1, len(S)):
        _L2 = [X | {x} for x in U - S[i] for X in L2]
        L2 = list(filter_minimal(_L2))

    return L2, U


def mbdllborder(isets1, isets2):
    """
    References
    ----------
    .. [1]
        Dong, Li
        Efficient Mining of Emerging Patterns Discovering

    Notes
    -----
    main algorithm, as explained in section 4.2
    """
    borders = list()

    for iset in isets2:
        if any((e > iset for e in isets1)):
            continue
        inter = (iset & e for e in isets1)
        R = filter_maximal(inter)

        diff = border_diff(iset, R)
        borders.append(diff)

    return borders


def borders_to_patterns(left, right, min_size=None):
    """
    Operates in a bread-first manner, outputting all
    valid patterns of a given size for each level.
    Bigger patterns first.

    Parameters
    ----------
    left: list of set
    right: set
    min_size: int
        only accepts patterns with greater or equal size

    Returns
    -------
    pd.Series
    """
    min_size = min_size or min(map(len, left))
    patterns = list()
    for size in range(len(right) - 1, min_size - 1, -1):  # bigger patterns first
        combs = combinations(right, size)
        for pat in combs:
            if any((e.issuperset(set(pat)) for e in left)):
                continue
            patterns.append(pat)

    return pd.Series(patterns)


class MBDLLBorder(BaseMiner, DiscovererMixin):
    """
    MBD-LLBorder aims at discovering patterns characterizing the difference
    between two collections of data.

    It first discovers ``two sets of maximal itemsets``, one for each collection.
    It then looks for borders of these sets, and characterizes the difference
    by only manipulating theses borders.

    This results in the algorithm only keeping a ``concise description (borders)``
    as an internal state.

    Last but not least, it enumerates the set of emerging patterns from the
    borders.

    Parameters
    ----------
    min_growth_rate: int or float, default=2
        A pattern is considered as emerging iff its support in the first collection
        is at least min_growth_rate times its support in the second collection.

    min_supp: int or float, default=.1
        Minimum support in each of the collection
        Must be a relative support between 0 and 1
        Default to 0.1 (10%)

    n_jobs : int, default=1
        The number of jobs to use for the computation.


    Attributes
    ----------
    borders_: list of tuple[list[set], set]
        List of pairs representing left and right borders
        For every pair, emerging patterns will be uncovered by enumerating itemsets
        from the right side, while checking for non membership in the left side


    References
    ----------
    .. [1]
        Guozhu Dong, Jinyan Li
        "Efficient Mining of Emerging Patterns : Discovering Trends and Differences", 1999
    """

    def __init__(self, min_growth_rate=2, min_supp=0.1, n_jobs=1):
        self.min_supp = _check_min_supp(min_supp, accept_absolute=False)
        self.min_growth_rate = _check_growth_rate(min_growth_rate)
        self.borders_ = None
        self.n_jobs = n_jobs

    def fit(self, D, y):
        """
        fit MBD-LLBorder on D, splitted on y

        This is done in two steps
            1. Discover maximal itemsets for the two disinct collections contained in D
            2. Mine borders from these maximal itemsets

        Parameters
        ----------
        D : pd.Series
            The input transactional database
            Where every entry contains singular items
            Items must be both hashable and comparable

        y : array-like of shape (n_samples,)
            Targets on which to split D
            Must contain only two disctinct values, i.e len(np.unique(y)) == 2
        """
        labels = np.unique(y)
        assert len(labels) == 2
        assert isinstance(D, pd.Series)  # TODO : accept tabular data

        D1, D2 = D[y == labels[0]], D[y == labels[1]]

        # TODO : replace LCMMax by some more efficient method
        right_border_d1 = LCMMax(min_supp=self.min_supp).fit_discover(D1)
        right_border_d2 = LCMMax(
            min_supp=self.min_growth_rate * self.min_supp
        ).fit_discover(D2)

        right_border_d1 = right_border_d1.itemset.map(set).tolist()
        right_border_d2 = right_border_d2.itemset.map(set).tolist()

        self.borders_ = mbdllborder(right_border_d1, right_border_d2)

        return self

    def discover(self, min_size=3):
        """
        Enumerate emerging patterns from borders
        Subsets are drawn from the right borders, and are accepted iff
        they do not belong to the corresponding left border.

        This implementation is parallel, we consider every couple
        of right/left border in a separate worker and run the computation

        Parameters
        ----------
        min_size: int
            minimum size for an itemset to be valid
        """
        btp = delayed(partial(borders_to_patterns, min_size=min_size))
        series = Parallel(n_jobs=self.n_jobs, prefer="processes")(
            btp(L, R) for L, R in self.borders_
        )
        return pd.concat(series) if series else pd.Series(dtype="object")
